"""Module to organize fluids."""
from warnings import warn

import CoolProp.CoolProp as CP


class Fluid:
    def __init__(self, name):
        self.name = name
        self.possible_names = []

    def __repr__(self):
        return f"{type(self).__name__}({self.__dict__}"


_fluid_list = CP.get_global_param_string("fluids_list").split(",")
fluid_list = {name: Fluid(name) for name in _fluid_list}

# define possible names
fluid_list["IsoButane"].possible_names.extend(
    ["isobutane", "i-butane", "ibutane", "isobutan", "iso-butane"]
)
fluid_list["n-Butane"].possible_names.extend(["butane", "n-butane", "nbutane"])
fluid_list["trans-2-Butene"].possible_names.extend(["trans-butene", "trans-butene-2"])
fluid_list["IsoButene"].possible_names.extend(["i-butene", "ibutene", "iso-butene"])
fluid_list["cis-2-Butene"].possible_names.extend(["cis-butene", "cis-butene-2"])
fluid_list["n-Pentane"].possible_names.extend(["pentane", "n-pentane", "npentane"])
fluid_list["Isopentane"].possible_names.extend(["i-pentane", "ipentane", "iso-pentane"])
fluid_list["n-Hexane"].possible_names.extend(["hexane", "n-hexane", "nhexane"])
fluid_list["Isohexane"].possible_names.extend(["isohexane", "i-hexane", "iso-hexane"])
fluid_list["n-Heptane"].possible_names.extend(["heptane", "n-heptane"])
fluid_list["n-Octane"].possible_names.extend(["octane", "n-octane"])
fluid_list["n-Undecane"].possible_names.extend(["undecane", "n-undecane"])
fluid_list["n-Dodecane"].possible_names.extend(["dodecane", "n-dodecane"])
fluid_list["HydrogenSulfide"].possible_names.extend(["hydrogen sulfide", "h2s"])
fluid_list["CarbonMonoxide"].possible_names.extend(["carbon monoxide", "co"])
fluid_list["CarbonDioxide"].possible_names.extend(["carbon dioxide"])
fluid_list["Nitrogen"].possible_names.extend(["n2"])
fluid_list["Oxygen"].possible_names.extend(["o2"])
fluid_list["Hydrogen"].possible_names.extend(["h2"])
fluid_list["Propylene"].possible_names.extend(["propene"])
fluid_list["Ethylene"].possible_names.extend(["ethene"])


def get_name(name):
    """Seach for compatible fluid name."""

    for k, v in fluid_list.items():
        if name.lower() in v.possible_names:
            name = k

    fluid_name = CP.get_REFPROPname(name)

    if fluid_name == "":
        raise ValueError(f"Fluid {name} not available. See ccp.fluid_list. ")

    return fluid_name


###############################################################################
# Helper functions
###############################################################################


def normalize_mix(molar_fractions):
    """
    Normalize the molar fractions so that the sum is 1.

    Parameters
    ----------
    molar_fractions : list
        Molar fractions of the components.

    Returns
    -------
    molar_fractions: list
        Molar fractions list will be modified in place.
    """
    total = sum(molar_fractions)
    if not ((0.95 < total < 1.05) or (95 < total < 105)):
        warn(f"Molar fraction far from 1 or 100% -> Total: {total}")

    for i, comp in enumerate(molar_fractions):
        molar_fractions[i] = comp / total

    try:
        # try to sum to exactly 1.
        molar_fractions_back = molar_fractions.copy()
        while sum(molar_fractions) != 1.0:
            diff = sum(molar_fractions) - 1.0
            molar_fractions[i] = molar_fractions[i] - diff
            i += 1
    except IndexError:
        molar_fractions = molar_fractions_back



from functools import reduce


def r_setattr(obj, attr, val):
    """Set attributes recursively."""
    # https://stackoverflow.com/a/31174427/5726899

    pre, _, post = attr.rpartition(".")
    return setattr(r_getattr(obj, pre) if pre else obj, post, val)


def r_getattr(obj, attr, *args):
    """Set attributes recursively, e.g.: r_getattr(curve, 'disch.p')."""

    def _getattr(obj, attr):
        return getattr(obj, attr, *args)

    return reduce(_getattr, [obj] + attr.split("."))



from copy import copy
from warnings import warn

import CoolProp.CoolProp as CP
import numpy as np
import ccp.config
from scipy.optimize import newton
from plotly import graph_objects as go
from . import _RP

from . import Q_
from .config.fluids import get_name, normalize_mix
from .config.units import check_units


class State(CP.AbstractState):
    """A thermodynamic state.

    .. note::
        To create a state use State.define(...)


    This class is inherited from CP.AbstractState.
    Some extra functionality has been added.


    Creates a state from fluid composition and two properties.
    Properties can be floats (SI units are considered) or pint quantities.

    Parameters
    ----------
    p : float, pint.Quantity
        Pressure
    T : float, pint.Quantity
        Temperature
    h : float, pint.Quantity
        Enthalpy
    s : float, pint.Quantity
        Entropy
    rho : float, pint.Quantity
        Specific mass

    fluid : dict
        Dictionary with constituent and composition.
        (e.g.: ({'Oxygen': 0.2096, 'Nitrogen': 0.7812, 'Argon': 0.0092})
    EOS : string
        String with HEOS or REFPROP

    Returns
    -------
    state : State object

    Examples
    --------
    >>> import ccp
    >>> Q_ = ccp.Q_
    >>> fluid = {'Oxygen': 0.2096, 'Nitrogen': 0.7812, 'Argon': 0.0092}
    >>> s = ccp.State.define(p=101008, T=273, fluid=fluid)
    >>> s.rho()
    <Quantity(1.28939426, 'kilogram / meter ** 3')>
    >>> # Using pint quantities
    >>> s = ccp.State.define(fluid=fluid, p=Q_(1, 'atm'), T=Q_(0, 'degC'))
    >>> s.h()
    <Quantity(273291.7, 'joule / kilogram')>
    """

    def __init__(self, EOS, _fluid):
        # no call to super(). see :
        # http://stackoverflow.com/questions/18260095/
        self.EOS = EOS
        self._fluid = _fluid

    def __repr__(self):
        args = {k: v for k, v in self.init_args.items() if v is not None}
        args_repr = [f'{k}=Q_("{getattr(self, k)():.0f~P}")' for k, v in args.items()]
        args_repr = ", ".join(args_repr)

        fluid_dict = self.fluid
        sorted_fluid_keys = sorted(fluid_dict, key=fluid_dict.get, reverse=True)
        fluid_repr = [f'"{k}": {fluid_dict[k]:.5f}' for k in sorted_fluid_keys]
        fluid_repr = "fluid={" + ", ".join(fluid_repr) + "}"

        return "State.define(" + args_repr + ", " + fluid_repr + ")"

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            self_fluid_rounded = {k: round(v, 3) for k, v in self.fluid.items()}
            other_fluid_rounded = {k: round(v, 3) for k, v in other.fluid.items()}
            if (
                self_fluid_rounded == other_fluid_rounded
                and np.allclose(self.p(), other.p(), rtol=1e-3)
                and np.allclose(self.T(), other.T(), rtol=1e-3)
            ):
                return True
        return False

    def _fluid_dict(self):
        # preserve the dictionary from define method
        fluid_dict = {}
        for k, v in zip(self.fluid_names(), self.get_mole_fractions()):
            fluid_dict[k] = v
            self.fluid = fluid_dict
        return fluid_dict

    def gas_constant(self, units=None):
        """Gas constant in joule / (mol kelvin).

        Returns
        -------
        gas_constant : pint.Quantity
            Gas constant (joule / (mol kelvin).
        """
        gas_constant = Q_(super().gas_constant(), "joule / (mol kelvin)")
        if units:
            gas_constant = gas_constant.to(units)
        return gas_constant

    def molar_mass(self, units=None):
        """Molar mass in kg/mol.

        Returns
        -------
        molar_mass : pint.Quantity
            Molar mass (kg/mol).
        """
        molar_mass = Q_(super().molar_mass(), "kg/mol")
        if units:
            molar_mass = molar_mass.to(units)
        return molar_mass

    def T(self, units=None):
        """Temperature in Kelvin.

        Returns
        -------
        T : pint.Quantity
            Temperature (Kelvin).
        """
        T = Q_(super().T(), "kelvin")
        if units:
            T = T.to(units)
        return T

    def p(self, units=None):
        """Pressure in Pascal.

        Returns
        -------
        p : pint.Quantity
            Pressure (pascal).
        """
        p = Q_(super().p(), "pascal")
        if units:
            p = p.to(units)
        return p

    def cp(self, units=None):
        """Specific heat at constant pressure joule/(kilogram kelvin).

        Returns
        -------
        cp : pint.Quantity
            Specific heat at constant pressure joule/(kilogram kelvin).
        """
        cp = Q_(super().cpmass(), "joule/(kilogram kelvin)")
        # use REFPROP directly with forced gas condition if cp value does not converge
        if cp < 0:
            fluids = self._fluid.replace("&", "*")
            r = _RP.REFPROPdll(
                fluids,
                "PTV",
                "Cp",
                _RP.MASS_BASE_SI,
                0,
                0,
                self.p("kPa").m,
                self.T().m,
                self.get_mole_fractions(),
            )
            cp = Q_(r.Output[0], "joule/(kilogram kelvin)")

        if units:
            cp = cp.to(units)
        return cp

    def cv(self, units=None):
        """Specific heat at constant volume joule/(kilogram kelvin).

        Returns
        -------
        cv : pint.Quantity
            Specific heat at constant volume joule/(kilogram kelvin).
        """
        cv = Q_(super().cvmass(), "joule/(kilogram kelvin)")
        if units:
            cv = cv.to(units)
        return cv

    def h(self, units=None):
        """Specific Enthalpy (joule/kilogram).

        Returns
        -------
        h : pint.Quantity
            Enthalpy (joule/kilogram).
        """
        h = Q_(super().hmass(), "joule/kilogram")
        if units:
            h = h.to(units)
        return h

    def s(self, units=None):
        """Specific entropy (per unit of mass).

        Returns
        -------
        s : pint.Quantity
            Entropy (joule/(kelvin kilogram)).
        """
        s = Q_(super().smass(), "joule/(kelvin kilogram)")
        if units:
            s = s.to(units)
        return s

    def p_critical(self, units=None):
        """Critical Pressure in Pa.

        Returns
        -------
        p_critical : pint.Quantity
            Critical pressure (Pa).
        """
        p_critical = Q_(super().p_critical(), "Pa")
        if units:
            p_critical = p_critical.to(units)
        return p_critical

    def T_critical(self, units=None):
        """Critical Temperature in K.

        Returns
        -------
        T_critical : pint.Quantity
            Critical temperature (degK).
        """
        T_critical = Q_(super().T_critical(), "K")
        if units:
            T_critical.to(units)
        return T_critical

    def rho(self, units=None):
        """Specific mass (kilogram/m**3).

        Returns
        -------
        rho : pint.Quantity
            Specific mass (kilogram/m**3).
        """
        rho = Q_(super().rhomass(), "kilogram/m**3")
        if units:
            rho = rho.to(units)
        return rho

    def v(self, units=None):
        """Specific volume (m**3/kilogram).

        Returns
        -------
        v : pint.Quantity
            Specific volume (m**3/kilogram).
        """
        v = 1 / self.rho()
        if units:
            v = (1 / self.rho()).to(units)
        return v

    def z(self, units=None):
        """Compressibility (dimensionless).

        Returns
        -------
        z : pint.Quantity
            Compressibility (dimensionless).
        """
        z = self.p() * self.molar_mass() / (self.rho() * self.gas_constant() * self.T())
        return z.to("dimensionless")

    def speed_sound(self, units=None):
        """Speed of sound - Eq. 8.1 from P. Nederstigt - Real Gas Thermodynamics.

        Returns
        -------
        speed_sound : pint.Quantity
            Speed of sound (m/s).
        """
        try:
            speed_sound = Q_(
                np.sqrt(self.first_partial_deriv(CP.iP, CP.iDmass, CP.iSmass)), "m/s"
            )
        except ValueError:
            # manually calculate the derivative for REFPROP 9.1
            dummy_state = copy(self)
            p0 = self.p()
            p1 = p0 + Q_(1e-6, "Pa")

            dummy_state.update(p=p1, s=self.s())
            rho0 = self.rho()
            rho1 = dummy_state.rho()
            delta_p = p1 - p0
            delta_rho = rho1 - rho0
            speed_sound = Q_(np.sqrt(delta_p / delta_rho), "m/s")

        if units:
            speed_sound = speed_sound.to(units)
        return speed_sound

    def viscosity(self, units=None):
        """Viscosity in pascal second.

        Returns
        -------
        viscosity : pint.Quantity
            Viscosity (pascal second)
        """
        try:
            viscosity = Q_(super().viscosity(), "pascal second")
        except ValueError:
            # handle error for cubic eos such as PR, SRK etc.
            dummy_state = self.define(
                p=self.p(), T=self.T(), fluid=self.fluid, EOS="REFPROP"
            )
            viscosity = Q_(super(State, dummy_state).viscosity(), "pascal second")
        if units:
            viscosity = viscosity.to(units)
        return viscosity

    def kinematic_viscosity(self, units=None):
        """Kinematic viscosity in m²/s.

        Returns
        -------
        kinematic_viscosity : pint.Quantity
            Kinematic viscosity (m²/s)
        """
        kinematic_viscosity = (self.viscosity() / self.rho()).to("m²/s")
        if units:
            kinematic_viscosity = kinematic_viscosity.to(units)
        return kinematic_viscosity

    def dpdv_s(self, units=None):
        """
        Partial derivative of pressure to spec. volume with const. entropy.
        """
        try:
            # dp/dv calculated from dp/drho needs to be multiplied by -rho**2
            dpdv_s = Q_(
                -self.rho().magnitude ** 2
                * (self.first_partial_deriv(CP.iP, CP.iDmass, CP.iSmass)),
                "pascal * kg / m**3",
            )
        except ValueError:
            # manually calculate the derivative for REFPROP 9.1
            dummy_state = copy(self)
            p0 = self.p()
            p1 = p0 + Q_(1e-1, "Pa")

            dummy_state.update(p=p1, s=self.s())
            v0 = self.v()
            v1 = dummy_state.v()
            dp = p1 - p0
            dv = v1 - v0
            dpdv_s = dp / dv
        if units:
            dpdv_s = dpdv_s.to(units)
        return dpdv_s

    def _X(self):
        """Coeficiente de compressibilidade X de Schultz"""
        T = self.T().to("K").magnitude
        V = self.v().to("m³/kg").magnitude

        return Q_(
            -T * V * self.first_partial_deriv(CP.iDmass, CP.iT, CP.iP) - 1,
            "dimensionless",
        )

    def _Y(self):
        """Coeficiente de compressibilidade X de Schultz"""
        P = self.p().to("Pa").magnitude
        V = self.v().to("m³/kg").magnitude

        return Q_(
            -(-P * V * self.first_partial_deriv(CP.iDmass, CP.iP, CP.iT)),
            "dimensionless",
        )

    def kv(self):
        """Isentropic volume exponent (dimensionless).

        Returns
        -------
        kv : pint.Quantity
            Isentropic volume exponent (dimensionless).
        """
        return -(self.v() / self.p()) * self.dpdv_s()

    def dTdp_s(self, units=None):
        """(dT / dp)s

        First partial derivative of temperature related to pressure with
        constant entropy."""
        try:
            dTdp_s = Q_(
                super().first_partial_deriv(CP.iT, CP.iP, CP.iSmass), "kelvin / pascal"
            )
        except ValueError:
            # manually calculate the derivative for REFPROP 9.1
            dummy_state = copy(self)
            p0 = self.p()
            p1 = p0 + Q_(1e-1, "Pa")

            dummy_state.update(p=p1, s=self.s())
            T0 = self.T()
            T1 = dummy_state.T()
            dp = p1 - p0
            dT = T1 - T0
            dTdp_s = dT / dp

        if units:
            dTdp_s = dTdp_s.to(units)

        return dTdp_s

    def kT(self):
        """Isentropic temperature exponent (dimensionless).

        Returns
        -------
        kT : pint.Quantity
            Isentropic temperature exponent (dimensionless).
        """
        return 1 / (1 - (self.p() / self.T()) * self.dTdp_s())

    def conductivity(self, units=None):
        """Thermal conductivity (W/m/K).

        Returns
        -------
        conductivity : pint.Quantity
            Thermal conductivity (W/m/K).
        """
        conductivity = Q_(super().conductivity(), "W/m/degK")
        if units:
            conductivity = conductivity.to(units)
        return conductivity

    def __reduce__(self):
        kwargs = dict(p=self.p(), T=self.T(), fluid=self.fluid)
        return self._rebuild, (self.__class__, kwargs)

    @staticmethod
    def _rebuild(cls, kwargs):
        return cls.define(**kwargs)

    @classmethod
    @check_units
    def define(
        cls,
        p=None,
        T=None,
        h=None,
        s=None,
        rho=None,
        fluid=None,
        EOS=None,
        **kwargs,
    ):
        """Constructor for state.

        Creates a state from fluid composition and two properties.
        Properties can be floats (SI units are considered) or pint quantities.

        Parameters
        ----------
        p : float, pint.Quantity
            Pressure
        T : float, pint.Quantity
            Temperature
        h : float, pint.Quantity
            Enthalpy
        s : float, pint.Quantity
            Entropy
        rho : float, pint.Quantity
            Specific mass

        fluid : dict
            Dictionary with constituent and composition.
            (e.g.: ({'Oxygen': 0.2096, 'Nitrogen': 0.7812, 'Argon': 0.0092})
        EOS : string
            String with HEOS or REFPROP

        Returns
        -------
        state : State object

        Examples
        --------
        >>> import ccp
        >>> Q_ = ccp.Q_
        >>> fluid = {'Oxygen': 0.2096, 'Nitrogen': 0.7812, 'Argon': 0.0092}
        >>> s = ccp.State.define(p=101008, T=273, fluid=fluid)
        >>> s.rho()
        <Quantity(1.28939426, 'kilogram / meter ** 3')>
        >>> # Using pint quantities
        >>> s = ccp.State.define(fluid=fluid, p=Q_(1, 'atm'), T=Q_(0, 'degC'))
        >>> s.h()
        <Quantity(273291.7, 'joule / kilogram')>
        """
        if fluid is None:
            raise TypeError("A fluid is required. Provide as fluid=dict(...)")

        if EOS is None:
            EOS = ccp.config.EOS

        constituents = []
        molar_fractions = []

        for k, v in fluid.items():
            k = get_name(k)
            constituents.append(k)
            molar_fractions.append(v)
            # create an adequate fluid string to cp.AbstractState
        _fluid = "&".join(constituents)

        state = cls(EOS, _fluid)

        normalize_mix(molar_fractions)
        state.set_mole_fractions(molar_fractions)
        state.init_args = dict(p=p, T=T, h=h, s=s, rho=rho)
        state.setup_args = copy(state.init_args)
        state.fluid = state._fluid_dict()
        if isinstance(fluid, str) and len(state.fluid) == 1:
            state.fluid[get_name(fluid)] = 1.0

        if isinstance(fluid, dict):
            if len(state.fluid) < len(fluid):
                raise ValueError(
                    "You might have repeated components in the fluid dictionary."
                )

        state.update(**state.setup_args)

        return state

    @check_units
    def update(
        self,
        p=None,
        T=None,
        rho=None,
        h=None,
        s=None,
        **kwargs,
    ):
        """Update the state.

        This method simplifies the state update. Only keyword arguments are
        required to update.

        Parameters
        ----------
        p : float, pint.Quantity
            Pressure (Pa).
        T : float, pint.Quantity
            Temperature (degk).
        rho : float, pint.Quantity
            Specific mass (kg/m**3).
        h : float, pint.Quantity
            Enthalpy (J/kg).
        s : float, pint.Quantity
            Entropy (J/(kg*degK)).
        """
        if p is not None and T is not None:
            super().update(CP.PT_INPUTS, p.magnitude, T.magnitude)
        elif p is not None and rho is not None:
            super().update(CP.DmassP_INPUTS, rho.magnitude, p.magnitude)
        elif p is not None and h is not None:
            super().update(CP.HmassP_INPUTS, h.magnitude, p.magnitude)
        elif p is not None and s is not None:
            if ccp.config.EOS == "REFPROP":
                super().update(CP.PSmass_INPUTS, p.magnitude, s.magnitude)
            else:
                # ps update not available for some EOS, this is a workaround based on:
                # https://github.com/CoolProp/CoolProp/issues/2000
                def objective(T):
                    super(State, self).update(CP.PT_INPUTS, p.magnitude, T)
                    return self.smass() - s.magnitude

                T0 = self.T().m
                if T0 == float("-inf"):
                    T0 = 300
                newton(objective, x0=T0)
        elif rho is not None and s is not None:
            try:
                super().update(CP.DmassSmass_INPUTS, rho.magnitude, s.magnitude)
            except ValueError:
                # handle convergence error by forcing gas state directly with REFPROP
                # calculate with p and T and update with their values
                fluids = self._fluid.replace("&", "*")
                r = _RP.REFPROPdll(
                    fluids,
                    "DSV",
                    "P,T",
                    _RP.MASS_BASE_SI,
                    0,
                    0,
                    rho.magnitude,
                    s.magnitude,
                    self.get_mole_fractions(),
                )
                super().update(CP.PT_INPUTS, r.Output[0], r.Output[1])
        elif rho is not None and T is not None:
            super().update(CP.DmassT_INPUTS, rho.magnitude, T.magnitude)
        elif h is not None and s is not None:
            super().update(CP.HmassSmass_INPUTS, h.magnitude, s.magnitude)
        elif T is not None and s is not None:
            super().update(CP.SmassT_INPUTS, s.magnitude, T.magnitude)
        else:
            locs = locals()
            for item in ["kwargs", "self", "__class__"]:
                locs.pop(item)
            raise KeyError(
                f"Update key "
                f"{[k for k, v in locs.items() if v is not None]}"
                f" not implemented"
            )

    def plot_envelope(
        self, T_units="degK", p_units="Pa", dew_point_margin=20, fig=None, **kwargs
    ):
        """Plot phase envelope.

        Plots the phase envelope and dew point limit.

        Parameters
        ----------
        T_units : str
            Temperature units.
            Default is 'degK'.
        p_units : str
            Pressure units.
            Default is 'Pa'.
        dew_point_margin : float
            Dew point margin.
            Default is 20 degK (from API). Unit is the same as T_units.
        fig : plotly.graph_objects.Figure, optional
            The figure object with the rotor representation.

        Returns
        -------
        fig : plotly.graph_objects.Figure
            The figure object with the rotor representation.
        """
        if fig is None:
            fig = go.Figure()

        if len(self.fluid) < 2:
            warn(
                "Pure fluids are not fully supported and might break things (e.g. plot_phase_envelope"
                "See https://github.com/CoolProp/CoolProp/issues/1544"
            )

        self.build_phase_envelope("dummy")
        phase_envelope = self.get_phase_envelope_data()
        T = Q_(np.array(phase_envelope.T), "degK").to(T_units).m
        p = Q_(np.array(phase_envelope.p), "Pa").to(p_units).m

        p_lower_bound = Q_(0.1, "atm").to(p_units).m
        T = T[p > p_lower_bound]
        p = p[p > p_lower_bound]

        T_dew = (
            np.add(
                T[: np.argmax(T)], np.multiply(dew_point_margin, np.ones(np.argmax(T)))
            ),
        )
        p_dew = (p[: np.argmax(T)],)

        hovertemplate = (
            f"Temperature ({T_units}): %{{x}}<br>" f"Pressure ({p_units}): %{{y}}"
        )

        fig.add_trace(
            go.Scatter(
                x=T,
                y=p,
                mode="lines",
                hovertemplate=hovertemplate,
                name="Phase Envelope",
            )
        )

        fig.add_trace(
            go.Scatter(
                x=T_dew[0],
                y=p_dew[0],
                mode="lines",
                line=dict(dash="dash"),
                hovertemplate=hovertemplate,
                name=f"Dew Point Margin ({dew_point_margin} {T_units})",
            )
        )

        fig.add_trace(
            go.Scatter(
                x=[self.T().to(T_units).m],
                y=[self.p().to(p_units).m],
                hovertemplate=hovertemplate,
                name="State",
            )
        )

        fig.update_layout(
            xaxis=dict(title_text=f"Temperature ({T_units})"),
            yaxis=dict(
                type="log", exponentformat="e", title_text=f"Pressure ({p_units})"
            ),
        )

        return fig

    def plot_point(self, T_units="degK", p_units="Pa", fig=None, **kwargs):
        """Plot point.

        Plot point in the given figure. Function will check for axis units and
        plot the point accordingly.

        Parameters
        ----------
        T_units : str
            Temperature units.
            Default is 'degK'.
        p_units : str
            Pressure units.
            Default is 'Pa'.
        fig : plotly.graph_objects.Figure, optional
            The figure object with the rotor representation.
        kwargs : dict
            Dictionary that will be passed to go.Scatter method.

        Returns
        -------
        fig : plotly.graph_objects.Figure
            The figure object with the rotor representation.
        """
        if fig is None:
            fig = go.Figure()

        p = self.p().to(p_units)
        T = self.T().to(T_units)

        default_values = dict(name="State")

        for k, v in default_values.items():
            kwargs.setdefault(k, v)

        fig.add_trace(
            go.Scatter(
                x=[T.m],
                y=[p.m],
                hovertemplate=f"Temperature ({T_units}): %{{x}}<br>"
                f"Pressure ({p_units}): %{{y}}",
                **kwargs,
            )
        )

        return fig



import pytest
import ccp
from numpy.testing import assert_allclose
from ccp.point import *
from pathlib import Path
from tempfile import tempdir
import pickle

skip = False  # skip slow tests


@pytest.fixture
def suc_0():
    fluid = dict(CarbonDioxide=0.76064, Nitrogen=0.23581, Oxygen=0.00284)
    return State.define(p=Q_(1.839, "bar"), T=291.5, fluid=fluid)


@pytest.fixture
def disch_0():
    fluid = dict(CarbonDioxide=0.76064, Nitrogen=0.23581, Oxygen=0.00284)
    return State.define(p=Q_(5.902, "bar"), T=405.7, fluid=fluid)


def test_raises_no_b_D(suc_0, disch_0):
    with pytest.raises(ValueError) as ex:
        Point(suc=suc_0, disch=disch_0, flow_v=1, speed=1)
        assert "Arguments b and D" in str(ex.value)
    with pytest.raises(ValueError) as ex:
        Point(suc=suc_0, disch=disch_0, flow_v=1, speed=1, b=1)
        assert "Arguments b and D" in str(ex.value)
    with pytest.raises(ValueError) as ex:
        Point(suc=suc_0, disch=disch_0, flow_v=1, speed=1, D=1)
        assert "Arguments b and D" in str(ex.value)


@pytest.fixture
def point_disch_flow_v_speed_suc(suc_0, disch_0):
    point_disch_suc = Point(suc=suc_0, disch=disch_0, flow_v=1, speed=1, b=1, D=1)
    return point_disch_suc


def test_point_disch_flow_v_speed_suc(suc_0, disch_0, point_disch_flow_v_speed_suc):
    assert point_disch_flow_v_speed_suc.suc == suc_0
    assert point_disch_flow_v_speed_suc.disch == disch_0
    assert_allclose(point_disch_flow_v_speed_suc.flow_v, 1.0)
    assert_allclose(point_disch_flow_v_speed_suc.flow_m, 3.072307, rtol=1e-4)
    assert_allclose(point_disch_flow_v_speed_suc.speed, 1.0)
    assert_allclose(point_disch_flow_v_speed_suc.head, 82877.366038, rtol=1e-4)
    assert_allclose(point_disch_flow_v_speed_suc.eff, 0.797811, rtol=1e-4)
    assert_allclose(point_disch_flow_v_speed_suc.phi, 2.546479)
    assert_allclose(point_disch_flow_v_speed_suc.psi, 663018.928304)
    assert_allclose(point_disch_flow_v_speed_suc.volume_ratio, 2.304738, rtol=1e-4)
    assert_allclose(point_disch_flow_v_speed_suc.power, 319154.332272)


@pytest.fixture
def point_eff_phi_psi_suc_volume_ratio(suc_0):
    point_eff_phi_psi_suc_volume_ratio = Point(
        suc=suc_0,
        eff=0.797811,
        phi=2.546479,
        psi=663018.928304,
        volume_ratio=2.304738,
        b=1,
        D=1,
    )
    return point_eff_phi_psi_suc_volume_ratio


def test_point_eff_suc_volume_ratio(suc_0, disch_0, point_eff_phi_psi_suc_volume_ratio):
    assert point_eff_phi_psi_suc_volume_ratio.suc == suc_0
    assert point_eff_phi_psi_suc_volume_ratio.disch == disch_0
    assert_allclose(point_eff_phi_psi_suc_volume_ratio.flow_v, 1.0, rtol=1e-4)
    assert_allclose(point_eff_phi_psi_suc_volume_ratio.flow_m, 3.072307, rtol=1e-4)
    assert_allclose(point_eff_phi_psi_suc_volume_ratio.speed, 1.0, rtol=1e-4)
    assert_allclose(point_eff_phi_psi_suc_volume_ratio.head, 82876.226229, rtol=1e-4)
    assert_allclose(point_eff_phi_psi_suc_volume_ratio.eff, 0.797811, rtol=1e-4)
    assert_allclose(point_eff_phi_psi_suc_volume_ratio.phi, 2.546479, rtol=1e-4)
    assert_allclose(point_eff_phi_psi_suc_volume_ratio.psi, 663018.928304, rtol=1e-4)
    assert_allclose(
        point_eff_phi_psi_suc_volume_ratio.volume_ratio, 2.304738, rtol=1e-4
    )
    assert_allclose(point_eff_phi_psi_suc_volume_ratio.power, 319154.332272, rtol=1e-3)


@pytest.fixture
def point_eff_flow_v_head_speed_suc(suc_0):
    point_eff_flow_v_head_speed_suc = Point(
        suc=suc_0, head=82876.226229, eff=0.797811, flow_v=1, speed=1, b=1, D=1
    )
    return point_eff_flow_v_head_speed_suc


def test_point_eff_flow_v_head_speed_suc(
    suc_0, disch_0, point_eff_flow_v_head_speed_suc
):
    assert point_eff_flow_v_head_speed_suc.suc == suc_0
    assert point_eff_flow_v_head_speed_suc.disch == disch_0
    assert_allclose(point_eff_flow_v_head_speed_suc.flow_v, 1.0, rtol=1e-4)
    assert_allclose(point_eff_flow_v_head_speed_suc.flow_m, 3.072307, rtol=1e-4)
    assert_allclose(point_eff_flow_v_head_speed_suc.speed, 1.0, rtol=1e-4)
    assert_allclose(point_eff_flow_v_head_speed_suc.head, 82876.226229, rtol=1e-4)
    assert_allclose(point_eff_flow_v_head_speed_suc.eff, 0.797811, rtol=1e-4)
    assert_allclose(point_eff_flow_v_head_speed_suc.phi, 2.546479, rtol=1e-4)
    assert_allclose(point_eff_flow_v_head_speed_suc.psi, 663009.809832, rtol=1e-4)
    assert_allclose(point_eff_flow_v_head_speed_suc.volume_ratio, 2.304715, rtol=1e-4)
    assert_allclose(point_eff_flow_v_head_speed_suc.power, 319149.832746, rtol=1e-4)


@pytest.fixture
def point_eff_flow_m_head_speed_suc(suc_0):
    point_eff_flow_v_head_speed_suc = Point(
        suc=suc_0, head=82876.226229, eff=0.797811, flow_m=3.072307, speed=1, b=1, D=1
    )
    return point_eff_flow_v_head_speed_suc


def test_point_eff_flow_m_head_speed_suc(
    suc_0, disch_0, point_eff_flow_m_head_speed_suc
):
    assert point_eff_flow_m_head_speed_suc.suc == suc_0
    assert point_eff_flow_m_head_speed_suc.disch == disch_0
    assert_allclose(point_eff_flow_m_head_speed_suc.flow_v, 1.0, rtol=1e-4)
    assert_allclose(point_eff_flow_m_head_speed_suc.flow_m, 3.072307, rtol=1e-4)
    assert_allclose(point_eff_flow_m_head_speed_suc.speed, 1.0, rtol=1e-4)
    assert_allclose(point_eff_flow_m_head_speed_suc.head, 82876.226229, rtol=1e-4)
    assert_allclose(point_eff_flow_m_head_speed_suc.eff, 0.797811, rtol=1e-4)
    assert_allclose(point_eff_flow_m_head_speed_suc.phi, 2.546479, rtol=1e-4)
    assert_allclose(point_eff_flow_m_head_speed_suc.psi, 663009.809832, rtol=1e-4)
    assert_allclose(point_eff_flow_m_head_speed_suc.volume_ratio, 2.304715, rtol=1e-4)
    assert_allclose(point_eff_flow_m_head_speed_suc.power, 319149.832746, rtol=1e-4)


@pytest.fixture
def point_disch_p_eff_flow_m_speed_suc(suc_0):
    point_disch_p_eff_flow_v_speed_suc = Point(
        suc=suc_0,
        disch_p=Q_(5.902, "bar"),
        eff=0.797811,
        flow_m=3.072307,
        speed=1,
        b=1,
        D=1,
    )
    return point_disch_p_eff_flow_v_speed_suc


def test_point_disch_p_eff_flow_m_head_speed_suc(
    suc_0, disch_0, point_disch_p_eff_flow_m_speed_suc
):
    assert point_disch_p_eff_flow_m_speed_suc.suc == suc_0
    assert point_disch_p_eff_flow_m_speed_suc.disch == disch_0
    assert_allclose(point_disch_p_eff_flow_m_speed_suc.flow_v, 1.0, rtol=1e-4)
    assert_allclose(point_disch_p_eff_flow_m_speed_suc.flow_m, 3.072307, rtol=1e-4)
    assert_allclose(point_disch_p_eff_flow_m_speed_suc.speed, 1.0, rtol=1e-4)
    assert_allclose(point_disch_p_eff_flow_m_speed_suc.head, 82876.226229, rtol=1e-4)
    assert_allclose(point_disch_p_eff_flow_m_speed_suc.eff, 0.797811, rtol=1e-4)
    assert_allclose(point_disch_p_eff_flow_m_speed_suc.phi, 2.546479, rtol=1e-4)
    assert_allclose(point_disch_p_eff_flow_m_speed_suc.psi, 663009.809832, rtol=1e-4)
    assert_allclose(
        point_disch_p_eff_flow_m_speed_suc.volume_ratio, 2.304715, rtol=1e-4
    )
    assert_allclose(point_disch_p_eff_flow_m_speed_suc.power, 319149.832746, rtol=1e-4)


@pytest.fixture
def point_disch_p_eff_flow_v_speed_suc(suc_0):
    point_disch_p_eff_flow_v_speed_suc = Point(
        suc=suc_0, disch_p=Q_(5.902, "bar"), eff=0.797811, flow_v=1, speed=1, b=1, D=1
    )
    return point_disch_p_eff_flow_v_speed_suc


def test_point_disch_p_eff_flow_v_head_speed_suc(
    suc_0, disch_0, point_disch_p_eff_flow_v_speed_suc
):
    assert point_disch_p_eff_flow_v_speed_suc.suc == suc_0
    assert point_disch_p_eff_flow_v_speed_suc.disch == disch_0
    assert_allclose(point_disch_p_eff_flow_v_speed_suc.flow_v, 1.0, rtol=1e-4)
    assert_allclose(point_disch_p_eff_flow_v_speed_suc.flow_m, 3.072307, rtol=1e-4)
    assert_allclose(point_disch_p_eff_flow_v_speed_suc.speed, 1.0, rtol=1e-4)
    assert_allclose(point_disch_p_eff_flow_v_speed_suc.head, 82876.226229, rtol=1e-4)
    assert_allclose(point_disch_p_eff_flow_v_speed_suc.eff, 0.797811, rtol=1e-4)
    assert_allclose(point_disch_p_eff_flow_v_speed_suc.phi, 2.546479, rtol=1e-4)
    assert_allclose(point_disch_p_eff_flow_v_speed_suc.psi, 663009.809832, rtol=1e-4)
    assert_allclose(
        point_disch_p_eff_flow_v_speed_suc.volume_ratio, 2.304715, rtol=1e-4
    )
    assert_allclose(point_disch_p_eff_flow_v_speed_suc.power, 319149.832746, rtol=1e-4)


def test_point_n_exp(suc_0, disch_0):
    assert_allclose(n_exp(suc_0, disch_0), 1.396545, rtol=1e-6)


def test_f_schultz(suc_0, disch_0):
    assert_allclose(f_schultz(suc_0, disch_0), 1.001647, rtol=1e-5)


def test_f_sandberg_colby(suc_0, disch_0):
    assert_allclose(f_sandberg_colby(suc_0, disch_0), 1.000912, rtol=1e-5)


def test_point_head_pol(suc_0, disch_0):
    h = head_pol(suc_0, disch_0)
    assert h.units == "joule/kilogram"
    assert_allclose(h, 82741.114339)


def test_head_pol_schultz(suc_0, disch_0):
    h = head_pol_schultz(suc_0, disch_0)
    assert h.units == "joule/kilogram"
    assert_allclose(h, 82877.366038, rtol=1e-6)


def test_head_pol_sandberg_colby(suc_0, disch_0):
    h = head_pol_sandberg_colby(suc_0, disch_0)
    assert h.units == "joule/kilogram"
    assert_allclose(h, 82816.596731, rtol=1e-6)


def test_point_head_pol_mallen_saville(suc_0, disch_0):
    h = head_pol_mallen_saville(suc_0, disch_0)
    assert h.units == "joule/kilogram"
    assert_allclose(h, 83006.348299)


def test_point_head_isen(suc_0, disch_0):
    h = head_isentropic(suc_0, disch_0)
    assert h.units == "joule/kilogram"
    assert_allclose(h.magnitude, 79984.234009, rtol=1e-5)


def test_head_reference(suc_0, disch_0):
    h, eff = head_reference(suc_0, disch_0)
    assert h.units == "joule/kilogram"
    assert_allclose(h, 82951.386575, rtol=1e-8)


def test_head_reference_2017(suc_0, disch_0):
    h, eff = head_reference_2017(suc_0, disch_0)
    assert h.units == "joule/kilogram"
    assert_allclose(h, 82951.388465, rtol=1e-8)


def test_head_pol_huntington(suc_0, disch_0):
    h = head_pol_huntington(suc_0, disch_0)
    assert h.units == "joule/kilogram"
    assert_allclose(h, 82951.470027, rtol=1e-6)


def test_point_eff_polytropic(suc_0, disch_0):
    assert_allclose(eff_pol(suc_0, disch_0), 0.796499, rtol=1e-5)


def test_point_eff_pol_schultz(suc_0, disch_0):
    assert_allclose(eff_pol_schultz(suc_0, disch_0), 0.797811, rtol=1e-5)


def test_point_eff_pol_huntington(suc_0, disch_0):
    assert_allclose(eff_pol_huntington(suc_0, disch_0), 0.798524, rtol=1e-5)


def test_eff_isentropic(suc_0, disch_0):
    assert_allclose(eff_isentropic(suc_0, disch_0), 0.76996, rtol=1e-5)


def test_reynolds(suc_0):
    re = reynolds(suc_0, speed=1, b=1, D=1)
    assert str(re.units) == "dimensionless"
    assert_allclose(re, 99944.204545)


def test_equality(point_disch_flow_v_speed_suc):
    point_1 = Point(
        suc=point_disch_flow_v_speed_suc.suc,
        speed=point_disch_flow_v_speed_suc.speed,
        flow_v=point_disch_flow_v_speed_suc.flow_v,
        head=point_disch_flow_v_speed_suc.head,
        eff=point_disch_flow_v_speed_suc.eff,
        b=point_disch_flow_v_speed_suc.b,
        D=point_disch_flow_v_speed_suc.D,
    )
    assert point_disch_flow_v_speed_suc == point_1


@pytest.fixture
def suc_1():
    fluid = dict(
        methane=0.69945,
        ethane=0.09729,
        propane=0.0557,
        nbutane=0.0178,
        ibutane=0.0102,
        npentane=0.0039,
        ipentane=0.0036,
        nhexane=0.0018,
        n2=0.0149,
        co2=0.09259,
        h2s=0.00017,
        water=0.002,
    )
    suc_1 = State.define(p=Q_(1.6995, "MPa"), T=311.55, fluid=fluid)

    return suc_1


@pytest.fixture
def point_eff_flow_v_head_speed_suc_1(suc_1):
    point_eff_flow_v_head_speed_suc_1 = Point(
        suc=suc_1,
        flow_v=Q_(6501.67, "m**3/h"),
        speed=Q_(11145, "RPM"),
        head=Q_(179.275, "kJ/kg"),
        eff=0.826357,
        b=Q_(28.5, "mm"),
        D=Q_(365, "mm"),
    )

    return point_eff_flow_v_head_speed_suc_1


def test_converted_from_find_speed(point_eff_flow_v_head_speed_suc_1):
    suc_2 = State.define(
        p=Q_(0.2, "MPa"), T=301.58, fluid={"n2": 1 - 1e-15, "co2": 1e-15}
    )
    point_converted_from_find_speed = Point.convert_from(
        original_point=point_eff_flow_v_head_speed_suc_1, suc=suc_2, find="speed"
    )

    assert_allclose(
        point_converted_from_find_speed.b, point_eff_flow_v_head_speed_suc_1.b
    )
    assert_allclose(
        point_converted_from_find_speed.D, point_eff_flow_v_head_speed_suc_1.D
    )
    assert_allclose(
        point_converted_from_find_speed.eff,
        point_eff_flow_v_head_speed_suc_1.eff,
        rtol=1e-4,
    )
    assert_allclose(
        point_converted_from_find_speed.phi,
        point_eff_flow_v_head_speed_suc_1.phi,
        rtol=1e-2,
    )
    assert_allclose(
        point_converted_from_find_speed.psi,
        point_eff_flow_v_head_speed_suc_1.psi,
        rtol=1e-2,
    )
    assert_allclose(point_converted_from_find_speed.volume_ratio_ratio, 1.0)
    assert_allclose(point_converted_from_find_speed.speed, 1259.934797)
    assert_allclose(point_converted_from_find_speed.head, 208933.668804, rtol=1e-2)
    assert_allclose(point_converted_from_find_speed.power, 1101698.5104, rtol=1e-2)


def test_converted_from_find_volume_ratio(point_eff_flow_v_head_speed_suc_1):
    suc_2 = State.define(
        p=Q_(0.2, "MPa"), T=301.58, fluid={"n2": 1 - 1e-15, "co2": 1e-15}
    )
    point_converted_from_find_volume_ratio = Point.convert_from(
        original_point=point_eff_flow_v_head_speed_suc_1, suc=suc_2, find="volume_ratio"
    )

    assert_allclose(
        point_converted_from_find_volume_ratio.b, point_eff_flow_v_head_speed_suc_1.b
    )
    assert_allclose(
        point_converted_from_find_volume_ratio.D, point_eff_flow_v_head_speed_suc_1.D
    )
    assert_allclose(
        point_converted_from_find_volume_ratio.eff,
        point_eff_flow_v_head_speed_suc_1.eff,
        rtol=1e-4,
    )
    assert_allclose(
        point_converted_from_find_volume_ratio.phi,
        point_eff_flow_v_head_speed_suc_1.phi,
        rtol=1e-2,
    )
    assert_allclose(
        point_converted_from_find_volume_ratio.psi,
        point_eff_flow_v_head_speed_suc_1.psi,
        rtol=1e-2,
    )
    assert_allclose(
        point_converted_from_find_volume_ratio.volume_ratio, 2.703027, rtol=1e-4
    )
    assert_allclose(point_converted_from_find_volume_ratio.speed, 1167.101671)
    assert_allclose(point_converted_from_find_volume_ratio.head, 179275.0, rtol=1e-2)
    assert_allclose(
        point_converted_from_find_volume_ratio.power, 875741.275802, rtol=1e-2
    )
    assert_allclose(point_converted_from_find_volume_ratio.phi_ratio, 1.0)
    assert_allclose(point_converted_from_find_volume_ratio.psi_ratio, 1.0)
    assert_allclose(
        point_converted_from_find_volume_ratio.volume_ratio_ratio, 0.882883, rtol=1e-4
    )


def test_converted_from_find_volume_ratio_mach_plot(point_eff_flow_v_head_speed_suc_1):
    suc_2 = State.define(
        p=Q_(0.2, "MPa"), T=301.58, fluid={"n2": 1 - 1e-15, "co2": 1e-15}
    )
    point_converted_from_find_volume_ratio = Point.convert_from(
        original_point=point_eff_flow_v_head_speed_suc_1, suc=suc_2, find="volume_ratio"
    )

    fig = point_converted_from_find_volume_ratio.plot_mach()

    assert_allclose(fig.data[2]["x"], 0.6012691126466259)
    assert_allclose(fig.data[2]["y"], 0.010087427961824047)


def test_converted_from_find_volume_ratio_reynolds_plot(
    point_eff_flow_v_head_speed_suc_1,
):
    suc_2 = State.define(
        p=Q_(0.2, "MPa"), T=301.58, fluid={"n2": 1 - 1e-15, "co2": 1e-15}
    )
    point_converted_from_find_volume_ratio = Point.convert_from(
        original_point=point_eff_flow_v_head_speed_suc_1, suc=suc_2, find="volume_ratio"
    )

    fig = point_converted_from_find_volume_ratio.plot_reynolds()

    assert_allclose(fig.data[2]["x"], 754805.237322)
    assert_allclose(fig.data[2]["y"], 0.091114, rtol=1e-5)


def test_save_load(point_disch_flow_v_speed_suc):
    file = Path(tempdir) / "suc_0.toml"
    point_disch_flow_v_speed_suc.save(file)
    point_0_loaded = Point.load(file)
    assert point_disch_flow_v_speed_suc == point_0_loaded


def test_pickle(point_disch_flow_v_speed_suc):
    pickled_point = pickle.loads(pickle.dumps(point_disch_flow_v_speed_suc))
    assert pickled_point == point_disch_flow_v_speed_suc
    assert hasattr(point_disch_flow_v_speed_suc, "head_plot") is True
    assert hasattr(pickled_point, "head_plot") is True


def test_global_polytropic_method(suc_0, disch_0):
    ccp.config.POLYTROPIC_METHOD = "huntington"
    p0 = Point(suc=suc_0, disch=disch_0, flow_v=1, speed=1, b=1, D=1)
    assert p0.head.units == "joule/kilogram"
    assert_allclose(p0.head, 82951.470027, rtol=1e-6)


def test_case_sc_at():
    suc = State.define(
        p=Q_("5516000 Pa"), T=Q_("311 K"), fluid={"CO2": 0.70000, "METHANE": 0.30000}
    )
    disch = State.define(
        p=Q_("16547000 Pa"), T=Q_("416 K"), fluid={"CO2": 0.70000, "METHANE": 0.30000}
    )
    assert_allclose(eff_pol_huntington(suc, disch), 0.818206, rtol=1e-6)



import pytest
import numpy as np
import pickle
from numpy.testing import assert_allclose
from ccp import ureg, Q_
from ccp.state import State
from ccp.point import Point
from ccp.curve import Curve


def test_raise_1_point():
    with pytest.raises(TypeError) as ex:
        Curve([1])
    assert "At least 2 points" in str(ex.value)


@pytest.fixture
def curve0():
    suc = State.define(p=Q_(1, "bar"), T=300, fluid={"co2": 1 - 1e-15, "n2": 1e-15})
    disch = State.define(p=Q_(2, "bar"), T=370, fluid={"co2": 1 - 1e-15, "n2": 1e-15})
    disch1 = State.define(
        p=Q_(2.5, "bar"), T=375, fluid={"co2": 1 - 1e-15, "n2": 1e-15}
    )
    p0 = Point(suc=suc, disch=disch, flow_v=1, speed=1, b=1, D=1)
    p1 = Point(suc=suc, disch=disch1, flow_v=2, speed=1, b=1, D=1)
    return Curve([p0, p1])


def test_curve_suc_parameters(curve0):
    assert curve0.suc.p().units == "pascal"
    assert_allclose(curve0.suc.p(), np.array([100000.0, 100000.0]))
    assert curve0.suc.T().units == "kelvin"
    assert_allclose(curve0.suc.T(), np.array([300.0, 300.0]))


def test_curve_disch_parameters(curve0):
    assert curve0.disch.p().units == "pascal"
    assert_allclose(curve0.disch.p(), np.array([200000.0, 250000.0]))
    assert curve0.disch.T().units == "kelvin"
    assert_allclose(curve0.disch.T(), np.array([370.0, 375.0]))


def test_curve_performance_parameters(curve0):
    assert curve0.head.units == "joule/kilogram"
    assert curve0.eff.units == ureg.dimensionless
    assert curve0.power.units == "watt"
    assert_allclose(curve0.head, np.array([43527.78012, 57942.686265]))
    assert_allclose(curve0.eff, np.array([0.709246, 0.881994]), rtol=1e-6)
    assert_allclose(curve0.power, np.array([108814.010351, 232958.372613]), rtol=1e-6)


def test_curve_interpolation(curve0):
    assert_allclose(curve0.disch.T_interpolated(1), 370.0)
    #  test for mutation of quantity magnitude
    a = Q_(1.0, "m**3/h")
    assert_allclose(curve0.disch.T_interpolated(a), 370.0)
    assert isinstance(a.m, float)


@pytest.fixture
def curve1():
    suc = State.define(p=Q_(1, "bar"), T=300, fluid={"co2": 1 - 1e-15, "n2": 1e-15})
    disch = State.define(p=Q_(2, "bar"), T=370, fluid={"co2": 1 - 1e-15, "n2": 1e-15})
    disch1 = State.define(
        p=Q_(2.5, "bar"), T=375, fluid={"co2": 1 - 1e-15, "n2": 1e-15}
    )
    disch2 = State.define(
        p=Q_(2.6, "bar"), T=376, fluid={"co2": 1 - 1e-15, "n2": 1e-15}
    )
    disch3 = State.define(
        p=Q_(2.7, "bar"), T=377, fluid={"co2": 1 - 1e-15, "n2": 1e-15}
    )
    p0 = Point(suc=suc, disch=disch, flow_v=1, speed=1, b=1, D=1)
    p1 = Point(suc=suc, disch=disch1, flow_v=2, speed=1, b=1, D=1)
    p2 = Point(suc=suc, disch=disch2, flow_v=3, speed=1, b=1, D=1)
    p3 = Point(suc=suc, disch=disch3, flow_v=4, speed=1, b=1, D=1)
    return Curve([p0, p1, p2, p3])


def test_pickle(curve0):
    pickled_curve0 = pickle.loads(pickle.dumps(curve0))
    assert pickled_curve0 == curve0
    assert hasattr(curve0, "head_plot") is True
    assert hasattr(pickled_curve0, "head_plot") is True



"""Module responsible for calculation.

This module will calculate and save json files with results.
"""
import ccp
import pandas as pd
import numpy as np
import schedule
import time
from datetime import datetime
from pathlib import Path

Q_ = ccp.Q_
CCP_PATH = Path(ccp.__file__).parent
DATA_PATH = CCP_PATH / "app/data"

# ----------------------
# Data
# ----------------------

df = pd.read_csv(
    CCP_PATH / "tests/data/UTGCA_1231_A_2019-01-01T00_00_00_2020-12-11T00_00_00_12h"
)
df = df.set_index("Unnamed: 0")
df.index.name = None


tags_a = {
    "Ts": "UTGCA_1231_TIT_218_A",  # Temperatura de sucção
    "ps": "UTGCA_1231_PIT_203_A",  # Pressão de sucção
    "Td": "UTGCA_1231_TIT_202_A",  # Temperatura de descarga
    "pd": "UTGCA_1231_PIT_204_A",  # Pressão de descarga
    "flow_v": "UTGCA_1231_FIT_201_A",  # Vazão
    "speed": "UTGCA_1231_SE_02_S_A",  # Rotação
}
tags_b = {
    "Ts": "UTGCA_1231_TIT_219_B",  # Temperatura de sucção
    "ps": "UTGCA_1231_PIT_207_B",  # Pressão de sucção
    "Td": "UTGCA_1231_TIT_205_B",  # Temperatura de descarga
    "pd": "UTGCA_1231_PIT_208_B",  # Pressão de descarga
    "flow_v": "UTGCA_1231_FIT_203_B",  # Vazão
    "speed": "UTGCA_1231_SE_02_S_B",  # Rotação
}
tags_c = {
    "Ts": "UTGCA_1231_TIT_220_C",  # Temperatura de sucção
    "ps": "UTGCA_1231_PIT_230_C",  # Pressão de sucção
    "Td": "UTGCA_1231_TIT_208_C",  # Temperatura de descarga
    "pd": "UTGCA_1231_PIT_212_C",  # Pressão de descarga
    "flow_v": "UTGCA_1231_FIT_205_C",  # Vazão
    "speed": "UTGCA_1231_SE_02_S_C",  # Rotação
}
tags_d = {
    "Ts": "UTGCA_1231_TIT_221_D",  # Temperatura de sucção
    "ps": "UTGCA_1231_PIT_224_D",  # Pressão de sucção
    "Td": "UTGCA_1231_TIT_213_D",  # Temperatura de descarga
    "pd": "UTGCA_1231_PIT_225_D",  # Pressão de descarga
    "flow_v": "UTGCA_1231_FIT_215_D",  # Vazão
    "speed": "UTGCA_1231_SE_02_S_D",  # Rotação
}
tags_e = {
    "Ts": "UTGCA_1231_TIT_222_E",  # Temperatura de sucção
    "ps": "UTGCA_1231_PIT_228_E",  # Pressão de sucção
    "Td": "UTGCA_1231_TIT_216_E",  # Temperatura de descarga
    "pd": "UTGCA_1231_PIT_229_E",  # Pressão de descarga
    "flow_v": "UTGCA_1231_FIT_216_E",  # Vazão
    "speed": "UTGCA_1231_SE_02_S_E",  # Rotação
}
tags_cromatografia = [
    # Cromatografia
    "UTGCA_1231_AI_002_C1",  # Analisador_METANO_AW-002',
    "UTGCA_1231_AI_002_C2",  # Analisador_ETANO_AW-002',
    "UTGCA_1231_AI_002_C3",  # CPR_Analisador_PROPANO_AW-002',
    "UTGCA_1231_AI_002_C6",  # Analisador_HEXANO_AW-002',
    "UTGCA_1231_AI_002_CO2",  # Analisador_GÁS CARBÔNICO_AW-002',
    "UTGCA_1231_AI_002_IC4",  # Analisador_I-BUTANO_AW-002',
    "UTGCA_1231_AI_002_IC5",  # Analisador_I-PENTANO_AW-002',
    "UTGCA_1231_AI_002_N2",  # Analisador_NITROGÊNIO_AW-002',
    "UTGCA_1231_AI_002_NC4",  # Analisador_N-BUTANO_AW-002',
    "UTGCA_1231_AI_002_NC5",  # Analisador_N-PENTANO_AW-002',
]
tags_dict = {
    "tags_a": tags_a,
    "tags_b": tags_b,
    "tags_c": tags_c,
    "tags_d": tags_d,
    "tags_e": tags_e,
}


class Data:
    def __init__(self, df):
        tags = ["a", "b", "c", "d", "e"]
        self.tags = tags

        for tag in tags:
            dfx = df[list(tags_dict[f"tags_{tag}"].values()) + tags_cromatografia]
            # eliminate errors such as 'comm fail'
            dfx = dfx.apply(pd.to_numeric, errors="coerce")
            # drop rows that have speed, flow etc. == 0.
            dfx = dfx[dfx != 0.0]
            dfx = dfx.dropna()
            setattr(self, f"df{tag}", dfx)


data = Data(df)


def get_imp_fd():
    composition_fd = dict(
        n2=0.4,
        co2=0.22,
        methane=92.11,
        ethane=4.94,
        propane=1.71,
        ibutane=0.24,
        butane=0.3,
        ipentane=0.04,
        pentane=0.03,
        hexane=0.01,
    )
    suc_fd = ccp.State.define(p=Q_(3876, "kPa"), T=Q_(11, "degC"), fluid=composition_fd)

    curve_name = "normal"
    curve_path = Path(CCP_PATH / "tests/data")

    imp_fd = ccp.Impeller.load_from_engauge_csv(
        suc=suc_fd,
        curve_name=curve_name,
        curve_path=curve_path,
        b=Q_(10.6, "mm"),
        D=Q_(390, "mm"),
        number_of_points=6,
        flow_units="kg/h",
        head_units="kJ/kg",
    )
    return imp_fd


def calculate_performance(tag, run_time, imp_fd):
    # TODO get data from PI
    # TODO organize data into required format
    data_tag = getattr(data, f"df{tag}")

    # mean around required time
    # calculate fluctuation -> percent difference between the minimum and
    # maximum reading divided by the average of all readings (PTC 10 table 3.4)
    # 3 readings on a timespan of 15 minutes.

    # get 3 random samples
    random_sample = np.random.randint(0, len(data_tag))
    data_tag = data_tag.iloc[random_sample - 1 : random_sample + 2]
    sample_time = data_tag.index[1]
    data_tag = data_tag.mean()

    tags = tags_dict[f"tags_{tag}"]
    ps_st = Q_(data_tag[tags["ps"]], "kPa")
    pd_st = Q_(data_tag[tags["pd"]], "kPa")
    Ts_st = Q_(data_tag[tags["Ts"]], "degC")
    Td_st = Q_(data_tag[tags["Td"]], "degC")
    flow_v_st = Q_(data_tag[tags["flow_v"]], "m³/h")
    speed_st = Q_(data_tag[tags["speed"]], "rpm")

    composition_st = dict(
        methane=data_tag.UTGCA_1231_AI_002_C1,  # Analisador_METANO_AW-002',
        ethane=data_tag.UTGCA_1231_AI_002_C2,  # Analisador_ETANO_AW-002',
        propane=data_tag.UTGCA_1231_AI_002_C3,  # CPR_Analisador_PROPANO_AW-002',
        hexane=data_tag.UTGCA_1231_AI_002_C6,  # Analisador_HEXANO_AW-002',
        co2=data_tag.UTGCA_1231_AI_002_CO2,  # Analisador_GÁS CARBÔNICO_AW-002',
        ibutane=data_tag.UTGCA_1231_AI_002_IC4,  # Analisador_I-BUTANO_AW-002',
        ipentane=data_tag.UTGCA_1231_AI_002_IC5,  # Analisador_I-PENTANO_AW-002',
        nitrogen=data_tag.UTGCA_1231_AI_002_N2,  # Analisador_NITROGÊNIO_AW-002',
        nbutane=data_tag.UTGCA_1231_AI_002_NC4,  # Analisador_N-BUTANO_AW-002',
        npentane=data_tag.UTGCA_1231_AI_002_NC5,  # Analisador_N-PENTANO_AW-002',
    )

    suc_st = ccp.State.define(p=ps_st, T=Ts_st, fluid=composition_st)
    disch_st = ccp.State.define(p=pd_st, T=Td_st, fluid=composition_st)
    imp_st = ccp.Impeller.convert_from(imp_fd, suc=suc_st)

    point_st = ccp.Point(
        speed=speed_st,
        flow_v=flow_v_st,
        suc=suc_st,
        disch=disch_st,
        b=Q_(10.6, "mm"),
        D=Q_(390, "mm"),
    )

    return imp_st, point_st, sample_time


def run_save_json(tag, run_time, imp_fd):
    run_time = datetime.now()
    print(f"Running {tag} : {run_time}")
    # erase old tags for that tag
    old_files = [f for f in DATA_PATH.glob("*.json")]
    old_files = [f for f in old_files if f"C-1231-{tag.capitalize()}" in str(f)]

    imp_op, point_op, sample_time = calculate_performance(tag, run_time, imp_fd)

    for curve in ["head", "eff", "power"]:
        # Head
        fig = getattr(imp_op, f"{curve}_plot")(
            flow_v=point_op.flow_v, speed=point_op.speed, speed_units="RPM"
        )
        fig = getattr(point_op, f"{curve}_plot")(
            fig=fig, speed_units="RPM", name="Operation Point"
        )

        fig.for_each_trace(
            lambda trace: trace.update(name="Expected Point")
            if "Flow" in trace.name
            else ()
        )
        run_time = str(run_time)[:19].replace(" ", "_").replace(":", "-")
        fig.write_json(
            str(
                DATA_PATH
                / f"C-1231-{tag.capitalize()}_{curve}_plot-time-{run_time}.json"
            )
        )

    for f in old_files:
        f.unlink()
    print(f"Finished {tag} : {run_time}")


if __name__ == "__main__":
    imp_fd = get_imp_fd()
    # download data
    # run
    run_time = datetime.now()

    for tag in ["a", "b"]:
        schedule.every(2).minutes.do(run_save_json, tag, run_time, imp_fd)

    while True:
        schedule.run_pending()
        time.sleep(1)



"""Module to read points from csv files created with engauge.

The csv files should be generated with engauge with the following procedure:
Inside engauge digitizer:
 - Edit -> Paste as new
 - On Axis Point -> add 3 reference points
 - Settings -> Curves -> With 'New', add how many curves will be collected
 - Segment fill
 - Select the curve (e.g. Curve1)
 - Select the points
 - Select the next curve and points...
 - Settings -> Export Setup -> Select:
 - Raws X's and Y's ; All curves on each line
Files should be saved with the following convention:
    - <curve-name>-head.csv
    - <curve-name>-eff.csv
"""
import csv


def read_data_from_engauge_csv(file_path):
    """Generate curves dict from file_path."""
    curves = {}

    with open(str(file_path)) as csv_file:
        csv_reader = csv.reader(csv_file)
        for row in csv_reader:
            if row:
                if row[0] == "x":
                    current_curve = row[1]
                    curves[current_curve] = {"x": [], "y": []}
                else:
                    curves[current_curve]["x"].append(float(row[0].replace(",", ".")))
                    curves[current_curve]["y"].append(float(row[1].replace(",", ".")))

    return curves



import pytest
import pickle
import ccp
from ccp.state import *
from numpy.testing import assert_allclose


def test_state_possible_name():
    with pytest.raises(ValueError) as exc:
        State.define(p=100000, T=300, fluid={"fake_name": 0.5, "fake_name2": 0.5})
    assert "Fluid fake_name not available." in str(exc.value)


def test_state_define():
    with pytest.raises(TypeError) as exc:
        State.define(p=100000, T=300)
    assert "A fluid is required" in str(exc.value)

    state = State.define(p=100000, T=300, fluid={"Methane": 1 - 1e-15, "Ethane": 1e-15})
    assert state.p().units == "pascal"
    assert state.T().units == "kelvin"
    assert state.p().magnitude == 100000
    assert state.T().magnitude == 300
    assert_allclose(state.rhomass(), 0.6442542612980722, rtol=1e-7)


def test_eos():
    ccp.config.EOS = "REFPROP"
    state = State.define(p=100000, T=300, fluid={"Methane": 1 - 1e-15, "Ethane": 1e-15})
    assert state.p().units == "pascal"
    assert state.T().units == "kelvin"
    assert state.p().magnitude == 100000
    assert state.T().magnitude == 300
    assert_allclose(state.rhomass(), 0.6442542612980722, rtol=1e-7)

    ccp.config.EOS = "PR"
    state = State.define(p=100000, T=300, fluid={"Methane": 1 - 1e-15, "Ethane": 1e-15})
    assert state.p().units == "pascal"
    assert state.T().units == "kelvin"
    assert state.p().magnitude == 100000
    assert state.T().magnitude == 300
    assert_allclose(state.rhomass(), 0.6445687063978816, rtol=1e-7)

    ccp.config.EOS = "SRK"
    state = State.define(p=100000, T=300, fluid={"Methane": 1 - 1e-15, "Ethane": 1e-15})
    assert state.p().units == "pascal"
    assert state.T().units == "kelvin"
    assert state.p().magnitude == 100000
    assert state.T().magnitude == 300
    assert_allclose(state.rhomass(), 0.6442384800595821, rtol=1e-7)

    ccp.config.EOS = "HEOS"
    state = State.define(p=100000, T=300, fluid={"Methane": 1 - 1e-15, "Ethane": 1e-15})
    assert state.p().units == "pascal"
    assert state.T().units == "kelvin"
    assert_allclose(state.p().magnitude, 100000)
    assert state.T().magnitude == 300
    assert state.rhomass() == 0.6442581578304425
    assert_allclose(state.rhomass(), 0.6442581578304425, rtol=1e-7)

    ccp.config.EOS = "REFPROP"


def test_state_define_units():
    state = State.define(
        p=Q_(1, "bar"),
        T=Q_(300 - 273.15, "celsius"),
        fluid={"Methane": 1 - 1e-15, "Ethane": 1e-15},
    )

    assert state.p().units == "pascal"
    assert state.T().units == "kelvin"
    assert state.rho().units == "kilogram/meter**3"
    assert state.p().magnitude == 100000
    assert state.T().magnitude == 300
    assert_allclose(state.rhomass(), 0.6442542612980722, rtol=1e-7)


def test_state_define_units_mix():
    state = State.define(
        p=Q_(1, "bar"),
        T=Q_(300 - 273.15, "celsius"),
        fluid={"Methane": 0.5, "Ethane": 0.5},
    )

    assert state.fluid == {"METHANE": 0.5, "ETHANE": 0.5}

    assert state.gas_constant().units == "joule/(kelvin mole)"
    assert state.molar_mass().units == "kilogram/mole"
    assert state.p().units == "pascal"
    assert state.T().units == "kelvin"
    assert state.rho().units == "kilogram/meter**3"
    assert state.v().units == "meter**3/kilogram"
    assert state.z().units == ""
    assert state.h().units == "joule/kilogram"
    assert state.s().units == "joule/(kelvin kilogram)"
    assert state.dpdv_s().units == "kilogram*pascal/meter**3"
    assert state.kv().units == ""
    assert state.dTdp_s().units == "kelvin/pascal"
    assert state.kT().units == ""

    assert state.gas_constant().magnitude == 8.314491
    assert state.molar_mass().magnitude == 0.02305592
    assert state.p().magnitude == 100000
    assert state.T().magnitude == 300
    assert_allclose(state.rho().magnitude, 0.9280595769591103)
    assert_allclose(state.v().magnitude, (1 / 0.9280595769591103))
    assert_allclose(state.z().magnitude, 0.99597784424262)
    assert_allclose(state.h().magnitude, 755784.43407392, rtol=1e-5)
    assert_allclose(state.s().magnitude, 4805.332018156618, rtol=1e-5)
    assert_allclose(state.dpdv_s().magnitude, -114182.00416892, rtol=1e-5)
    assert_allclose(state.kv().magnitude, 1.230331, rtol=1e-5)
    assert_allclose(state.dTdp_s().magnitude, 0.5664135e-3, rtol=1e-5)
    assert_allclose(state.kT().magnitude, 1.232748, rtol=1e-5)
    assert (
        state.__repr__()
        == 'State.define(p=Q_("100000 Pa"), T=Q_("300 K"), fluid={"METHANE": 0.50000, "ETHANE": 0.50000})'
    )

    state.update(p=200000, T=310)
    assert state.p().units == "pascal"
    assert state.T().units == "kelvin"
    assert state.rho().units == "kilogram/meter**3"
    assert state.p().magnitude == 200000
    assert state.T().magnitude == 310
    assert_allclose(state.rho().magnitude, 1.8020813868455758)
    assert (
        state.__repr__()
        == 'State.define(p=Q_("200000 Pa"), T=Q_("310 K"), fluid={"METHANE": 0.50000, "ETHANE": 0.50000})'
    )


def test_state_copy():
    state = State.define(p=100000, T=300, fluid={"Methane": 1 - 1e-15, "Ethane": 1e-15})
    state1 = copy(state)

    assert state == state
    assert state.rho() == state1.rho()


def test_rho_p_inputs():
    state = State.define(
        rho=0.9280595769591103, p=Q_(1, "bar"), fluid={"Methane": 0.5, "Ethane": 0.5}
    )
    assert_allclose(state.T().m, 300)


def test_rho_T_inputs():
    state = State.define(
        rho=0.9280595769591103, T=300, fluid={"Methane": 0.5, "Ethane": 0.5}
    )
    assert_allclose(state.p().m, 100000)


def test_h_s_inputs():
    state = State.define(
        h=755784.43407392, s=4805.332018156618, fluid={"Methane": 0.5, "Ethane": 0.5}
    )
    assert_allclose(state.h().magnitude, 755784.43407392, rtol=1e-5)
    assert_allclose(state.s().magnitude, 4805.332018156618, rtol=1e-5)
    assert_allclose(state.rho().magnitude, 0.9280595769591103, rtol=1e-5)


def test_h_p_inputs():
    state = State.define(
        h=755784.43407392, p=Q_(1, "bar"), fluid={"Methane": 0.5, "Ethane": 0.5}
    )
    assert_allclose(state.h().magnitude, 755784.43407392, rtol=1e-5)
    assert_allclose(state.s().magnitude, 4805.332018156618, rtol=1e-5)
    assert_allclose(state.rho().magnitude, 0.9280595769591103)


def test_T_s_inputs():
    state = State.define(
        T=300, s=4805.332018156618, fluid={"Methane": 0.5, "Ethane": 0.5}
    )
    assert_allclose(state.h().magnitude, 755784.43407392, rtol=1e-5)
    assert_allclose(state.s().magnitude, 4805.332018156618, rtol=1e-5)
    assert_allclose(state.rho().magnitude, 0.9280595769591103, rtol=1e-5)


def test_equality():
    state = State.define(p=100000, T=300, fluid={"Methane": 1 - 1e-15, "Ethane": 1e-15})
    state1 = State.define(p=state.p(), T=state.T(), fluid=state.fluid)
    state2 = State.define(h=state.h(), s=state.s(), fluid=state.fluid)

    assert state == state1
    assert state == state2
    assert state1 == state2

    state_mix = State.define(p=100000, T=300, fluid={"Methane": 0.5, "Ethane": 0.5})
    state1_mix = State.define(p=state_mix.p(), T=state_mix.T(), fluid=state_mix.fluid)
    state2_mix = State.define(h=state_mix.h(), s=state_mix.s(), fluid=state_mix.fluid)

    assert state_mix == state1_mix
    assert state_mix == state2_mix
    assert state1_mix == state2_mix


def test_mix_composition():
    fluid = {
        "Isobutene": 0.20,
        "HYDROGEN SULFIDE": 2.67,
        "HEXANE": 7.01,
        "propylene": 0.55,
        "ISOBUTANE": 5.43,
        "Methane": 7.04,
        "ethylene": 0.24,
        "hydrogen": 0.75,
        "Nitrogen": 11.39,
        "BUTANE": 26.70,
        "PROPANE": 21.23,
        "ETHANE": 2.88,
        "1Butene": 0.16,
        "C2BUTENE": 0.02,
        "ISOPENTANE": 5.12,
        "PENTANE": 7.11,
        "T2BUTENE": 0.02,
        "CO": 0.03,
        "carbon dioxide": 1.15,
        "N2": 0.30,
    }

    with pytest.raises(ValueError) as exc:
        State.define(p=Q_(0.804, "kgf/cm**2"), T=Q_(37.4, "degC"), fluid=fluid)
    assert "You might have repeated" in str(exc.value)


def test_pickle():
    state = State.define(p=100000, T=300, fluid={"Methane": 1 - 1e-15, "Ethane": 1e-15})
    assert pickle.loads(pickle.dumps(state)) == state



from copy import copy

import numpy as np
import toml
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from scipy.optimize import newton

import ccp.config
from .state import State
from ccp.config.units import check_units, Q_
from ccp.config.utilities import r_getattr


class Point:
    """A performance point.
    A point in the compressor map that can be defined in different ways.

    Parameters
    ----------
    speed : pint.Quantity, float
        Speed in rad/s.
    flow_v or flow_m : pint.Quantity, float
        Volumetric (m³/s) or mass (kg/s) flow.
    suc, disch : ccp.State, ccp.State
        Suction and discharge states for the point.
    suc, disch_p, eff : ccp.State, float, float
        Suction state, discharge pressure and polytropic efficiency.
    suc, head, eff : ccp.State, float, float
        Suction state, polytropic head and polytropic efficiency.
    suc, head, power : ccp.State, pint.Quantity or float, pint.Quantity or float
        Suction state, polytropic head (J/kg) and gas power (Watt).
    suc, eff, volume_ratio : ccp.State, float, float
        Suction state, polytropic efficiency and volume ratio.
    b, D : pint.Quantity, float
        Impeller width and diameter.

    Returns
    -------
    Point : ccp.Point
        A point in the compressor map.

    Attributes
    ----------
    suc : ccp.State
        A ccp.State object.
        For more information on attributes and methods available see:
        :py:class:`ccp.State`
    disch : ccp.State
        A ccp.State object.
        For more information on attributes and methods available see:
        :py:class:`ccp.State`
    flow_v : pint.Quantity
        Volumetric flow (m³/s).
    flow_m : pint.Quantity
        Mass flow (kg/s)
    speed : pint.Quantity
        Speed (rad/s).
    head : pint.Quantity
        Polytropic head (J/kg).
    eff : pint.Quantity
        Polytropic efficiency (dimensionless).
    power : pint.Quantity
        Power (Watt).
    phi : pint.Quantity
        Volume flow coefficient (dimensionless).
    psi : pint.Quantity
        Polytropic head coefficient (dimensionless).
    volume_ratio : pint.Quantity
        Volume ratio - suc.v() / disch.v() (dimensionless).
    b : pint.Quantity
        Impeller width (m).
    D : pint.Quantity
        Impeller diameter (m).
    reynolds : pint.Quantity
        Reynolds number (dimensionless).
    mach : pint.Quantity
        Mach number (dimensionless).
    phi_ratio : float
        Ratio between phi for this point and the original point from which it was converted from.
    psi_ratio : float
        Ratio between psi for this point and the original point from which it was converted from.
    reynolds_ratio : float
        Ratio between Reynolds for this point and the original point from which it was converted from.
    mach_diff : float
        Difference between Mach for this point and the original point from which it was converted from.
    volume_ratio_ratio = 1.0
        Ratio between volume_ratio for this point and the original point from which it was converted from.
    polytropic_method : str, optional
        Polytropic method used for head and efficiency calculation.
        Options are: "mallen_saveille", "sandberg_colby", "schultz" and "huntington".
        The default is "schultz".
        The default value can be changed in a global level with:
        ccp.config.POLYTROPIC_METHOD = "<desired value>"
    """

    @check_units
    def __init__(
        self,
        suc=None,
        disch=None,
        disch_p=None,
        flow_v=None,
        flow_m=None,
        speed=None,
        head=None,
        eff=None,
        power=None,
        phi=None,
        psi=None,
        volume_ratio=None,
        b=None,
        D=None,
        polytropic_method=None,
    ):
        if b is None or D is None:
            raise ValueError("Arguments b and D must be provided.")
        if polytropic_method is None:
            polytropic_method = ccp.config.POLYTROPIC_METHOD

        self.head_calc_func = globals()[f"head_pol_{polytropic_method}"]
        self.eff_calc_func = globals()[f"eff_pol_{polytropic_method}"]

        self.suc = suc
        self.disch = disch
        self.disch_p = disch_p
        self.flow_v = flow_v
        self.flow_m = flow_m
        self.speed = speed
        self.head = head
        self.eff = eff
        self.power = power

        self.phi = phi
        self.psi = psi
        self.volume_ratio = volume_ratio

        self.b = b
        self.D = D

        # dummy state used to avoid copying states
        self._dummy_state = copy(self.suc)

        kwargs_list = []

        for k in [
            "suc",
            "disch",
            "disch_p",
            "flow_v",
            "flow_m",
            "speed",
            "head",
            "eff",
            "power",
            "phi",
            "psi",
            "volume_ratio",
        ]:
            if getattr(self, k) is not None:
                kwargs_list.append(k)

        kwargs_str = "_".join(sorted(kwargs_list))

        getattr(self, "_calc_from_" + kwargs_str)()

        self.reynolds = reynolds(self.suc, self.speed, self.b, self.D)
        self.mach = mach(self.suc, self.speed, self.D)

        self.phi_ratio = 1.0
        self.psi_ratio = 1.0
        self.reynolds_ratio = 1.0
        # mach in the ptc 10 is compared with Mmt - Mmsp
        self.mach_diff = 0.0
        # ratio between specific volume ratios in original and converted conditions
        self.volume_ratio_ratio = 1.0

        self._add_point_plot()

    def _add_point_plot(self):
        """Add plot to point after point is fully defined."""
        for state in ["suc", "disch"]:
            for attr in ["p", "T", "h", "s", "rho"]:
                plot = plot_func(self, ".".join([state, attr]))
                setattr(getattr(self, state), attr + "_plot", plot)
        for attr in ["head", "eff", "power"]:
            plot = plot_func(self, attr)
            setattr(self, attr + "_plot", plot)

    def __str__(self):
        return (
            f"\nPoint: "
            f"\nVolume flow: {self.flow_v:.2f~P}"
            f"\nHead: {self.head:.2f~P}"
            f"\nEfficiency: {self.eff:.2f~P}"
            f"\nPower: {self.power:.2f~P}"
        )

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if (
                self.suc == other.suc
                and np.allclose(self.speed, other.speed)
                and np.allclose(self.flow_v, other.flow_v)
                and np.allclose(self.head, other.head)
                and np.allclose(self.eff, other.eff)
            ):
                return True

        return False

    def __repr__(self):

        return (
            f"{self.__class__.__name__}(suc={self.suc},"
            f' speed=Q_("{self.speed:.0f~P}"),'
            f' flow_v=Q_("{self.flow_v:.2f~P}"),'
            f' head=Q_("{self.head:.0f~P}"),'
            f' eff=Q_("{self.eff:.3f~P}"))'
        )

    def _calc_from_disch_flow_v_speed_suc(self):
        self.head = self.head_calc_func(self.suc, self.disch)
        self.eff = self.eff_calc_func(self.suc, self.disch)
        self.volume_ratio = self.suc.v() / self.disch.v()
        self.flow_m = self.suc.rho() * self.flow_v
        self.power = power_calc(self.flow_m, self.head, self.eff)
        self.phi = phi(self.flow_v, self.speed, self.D)
        self.psi = psi(self.head, self.speed, self.D)

    def _calc_from_disch_flow_m_speed_suc(self):
        self.head = self.head_calc_func(self.suc, self.disch)
        self.eff = self.eff_calc_func(self.suc, self.disch)
        self.volume_ratio = self.suc.v() / self.disch.v()
        self.flow_v = self.flow_m / self.suc.rho()
        self.power = power_calc(self.flow_m, self.head, self.eff)
        self.phi = phi(self.flow_v, self.speed, self.D)
        self.psi = psi(self.head, self.speed, self.D)

    def _calc_from_eff_phi_psi_suc_volume_ratio(self):
        eff = self.eff
        suc = self.suc
        volume_ratio = self.volume_ratio

        disch_v = suc.v() / volume_ratio
        disch_rho = 1 / disch_v

        # consider first an isentropic compression
        disch = State.define(rho=disch_rho, s=suc.s(), fluid=suc.fluid)

        def update_state(x, update_type):
            if update_type == "pressure":
                disch.update(rho=disch_rho, p=x)
            elif update_type == "temperature":
                disch.update(rho=disch_rho, T=x)
            new_eff = self.eff_calc_func(self.suc, disch)
            if not 0.0 < new_eff < 1.1:
                raise ValueError

            return (new_eff - eff).magnitude

        try:
            newton(update_state, disch.T().magnitude, args=("temperature",), tol=1e-1)
        except ValueError:
            # re-instantiate disch, since update with temperature not converging
            # might break the state
            disch = State.define(rho=disch_rho, s=suc.s(), fluid=suc.fluid)
            newton(update_state, disch.p().magnitude, args=("pressure",), tol=1e-1)

        self.disch = disch
        self.head = self.head_calc_func(suc, disch)
        self.speed = speed_from_psi(self.D, self.head, self.psi)
        self.flow_v = flow_from_phi(self.D, self.phi, self.speed)
        self.flow_m = self.flow_v * self.suc.rho()
        self.power = power_calc(self.flow_m, self.head, self.eff)

    def _calc_from_eff_flow_v_head_speed_suc(self):
        eff = self.eff
        head = self.head
        suc = self.suc
        disch = disch_from_suc_head_eff(suc, head, eff)
        self.disch = disch
        self.flow_m = self.flow_v * self.suc.rho()
        self.power = power_calc(self.flow_m, self.head, self.eff)
        self.phi = phi(self.flow_v, self.speed, self.D)
        self.psi = psi(self.head, self.speed, self.D)
        self.volume_ratio = self.suc.v() / self.disch.v()

    def _calc_from_eff_flow_m_head_speed_suc(self):
        eff = self.eff
        head = self.head
        suc = self.suc
        disch = disch_from_suc_head_eff(suc, head, eff)
        self.disch = disch
        self.flow_v = self.flow_m / self.suc.rho()
        self.power = power_calc(self.flow_m, self.head, self.eff)
        self.phi = phi(self.flow_v, self.speed, self.D)
        self.psi = psi(self.head, self.speed, self.D)
        self.volume_ratio = self.suc.v() / self.disch.v()

    def _calc_from_eff_phi_psi_speed_suc(self):
        self.head = head_from_psi(self.D, self.psi, self.speed)
        self.disch = disch_from_suc_head_eff(self.suc, self.head, self.eff)
        self.flow_v = flow_from_phi(self.D, self.phi, self.speed)
        self.flow_m = self.flow_v * self.suc.rho()
        self.power = power_calc(self.flow_m, self.head, self.eff)
        self.volume_ratio = self.suc.v() / self.disch.v()

    def _calc_from_disch_p_eff_flow_v_speed_suc(self):
        eff = self.eff
        suc = self.suc
        disch = disch_from_suc_disch_p_eff(suc, self.disch_p, eff)
        self.disch = disch
        self.head = self.head_calc_func(suc, disch)
        self.flow_m = self.flow_v * self.suc.rho()
        self.power = power_calc(self.flow_m, self.head, self.eff)
        self.phi = phi(self.flow_v, self.speed, self.D)
        self.psi = psi(self.head, self.speed, self.D)
        self.volume_ratio = self.suc.v() / self.disch.v()

    def _calc_from_disch_p_eff_flow_m_speed_suc(self):
        eff = self.eff
        suc = self.suc
        disch = disch_from_suc_disch_p_eff(suc, self.disch_p, eff)
        self.disch = disch
        self.head = self.head_calc_func(suc, disch)
        self.flow_v = self.flow_m / self.suc.rho()
        self.power = power_calc(self.flow_m, self.head, self.eff)
        self.phi = phi(self.flow_v, self.speed, self.D)
        self.psi = psi(self.head, self.speed, self.D)
        self.volume_ratio = self.suc.v() / self.disch.v()

    @classmethod
    @check_units
    def convert_from(cls, original_point, suc=None, find="speed", speed=None):
        """Convert point from an original point.

        Parameters
        ----------
        original_point : ccp.Point
            Original point from which the desired point will be converted.
        find : str, optional
            If the calculation will find a new speed keeping constant volume ratio,
            or a new volume ratio for the desired speed.
            Options are "speed" or "volume_ratio", default is "speed".
        speed : float, pint.Quantity, optional
            Desired speed. If find="speed", this should be None.

        The user must provide 3 of the 4 available arguments. The argument which is not
        provided will be calculated.
        """
        if speed is None:
            speed = original_point.speed
        convert_point_options = {
            "speed": dict(
                suc=suc,
                eff=original_point.eff,
                phi=original_point.phi,
                psi=original_point.psi,
                volume_ratio=original_point.volume_ratio,
                b=original_point.b,
                D=original_point.D,
            ),
            "volume_ratio": dict(
                suc=suc,
                eff=original_point.eff,
                phi=original_point.phi,
                psi=original_point.psi,
                speed=speed,
                b=original_point.b,
                D=original_point.D,
            ),
        }

        converted_point = cls(**convert_point_options[find])
        converted_point.phi_ratio = converted_point.phi / original_point.phi
        converted_point.psi_ratio = converted_point.psi / original_point.psi
        converted_point.volume_ratio_ratio = (
            converted_point.volume_ratio / original_point.volume_ratio
        )
        converted_point.reynolds_ratio = (
            converted_point.reynolds / original_point.reynolds
        )
        converted_point.mach_diff = converted_point.mach - original_point.mach

        return converted_point

    def __getstate__(self):
        attributes = self.__dict__.copy()
        final_attributes = {k: v for k, v in attributes.items() if "plot" not in k}

        return final_attributes

    def __setstate__(self, state):
        self.__dict__ = state
        self._add_point_plot()

    def _dict_to_save(self):
        """Returns a dict that will be saved to a toml file."""
        return dict(
            p=str(self.suc.p()),
            T=str(self.suc.T()),
            fluid=self.suc.fluid,
            speed=str(self.speed),
            flow_v=str(self.flow_v),
            head=str(self.head),
            eff=str(self.eff),
            b=str(self.b),
            D=str(self.D),
        )

    @staticmethod
    def _dict_from_load(dict_parameters):
        """Change dict to format that can be used by load constructor."""
        suc = State.define(
            p=Q_(dict_parameters.pop("p")),
            T=Q_(dict_parameters.pop("T")),
            fluid=dict_parameters.pop("fluid"),
        )

        return dict(suc=suc, **{k: Q_(v) for k, v in dict_parameters.items()})

    def save(self, file_name):
        """Save point to toml file."""
        with open(file_name, mode="w") as f:
            toml.dump(self._dict_to_save(), f)

    @classmethod
    def load(cls, file_name):
        """Load point from toml file."""
        with open(file_name) as f:
            parameters = toml.load(f)

        return cls(**cls._dict_from_load(parameters))

    def mach_limits(self, mmsp=None):
        """Calculate Mach lower and upper limits.

        Parameters
        ----------
        mmsp : float, optional
            Mach number specified. Default value is the point Mach number.

        Returns
        -------
        limits : dict
            Dict with keys: 'lower', 'upper' and 'within_limits'.
        """
        if mmsp is None:
            mmsp = self.mach.m
        if 0 <= mmsp < 0.214:
            lower_limit = -mmsp
            upper_limit = -0.25 * mmsp + 0.286
        elif 0.214 <= mmsp < 0.86:
            lower_limit = 0.266 * mmsp - 0.271
            upper_limit = -0.25 * mmsp + 0.286
        elif mmsp >= 0.86:
            lower_limit = -0.042
            upper_limit = 0.07
        else:
            raise ValueError("Mach number out of specified range.")

        if lower_limit < self.mach_diff < upper_limit:
            within_limits = True
        else:
            within_limits = False

        return {
            "lower": lower_limit,
            "upper": upper_limit,
            "within_limits": within_limits,
        }

    def plot_mach(self, fig=None, **kwargs):
        """Plot allowable Mach range and point.

        This will plot the allowable Mach range and the point according to the
        PTC criteria.

        Parameters
        ----------
        fig : plotly.Figure
            Plotly figure.

        Returns
        -------
        fig : plotly.Figure
            Plotly figure.
        """
        if fig is None:
            fig = go.Figure()

        # build acceptable region
        upper_limit = []
        lower_limit = []
        mmsp_range = np.linspace(0, 1.6, 300)
        for mmsp in mmsp_range:
            mach_limits = self.mach_limits(mmsp)
            lower_limit.append(mach_limits["lower"])
            upper_limit.append(mach_limits["upper"])

        fig.add_trace(
            go.Scatter(x=mmsp_range, y=lower_limit, marker=dict(color="black"))
        )
        fig.add_trace(
            go.Scatter(x=mmsp_range, y=upper_limit, marker=dict(color="black"))
        )

        # add point
        fig.add_trace(
            go.Scatter(x=[self.mach.m], y=[self.mach_diff], marker=dict(color="red"))
        )

        fig.update_xaxes(
            title=r"$\text{Machine Mach No. Specified} - Mm_{sp}$",
        )
        fig.update_yaxes(
            title=r"$Mm_t - Mm_{sp}$",
        )
        fig.update_layout(showlegend=False)

        return fig

    def reynolds_limits(self, remsp=None):
        """Calculate Reynolds lower and upper limits.

        Parameters
        ----------
        remsp : float, optional
            Reynolds number specified. Default value is the point reynolds number.

        Returns
        -------
        limits : dict
            Dict with keys: 'lower', 'upper' and 'within_range'.
        """
        if remsp is None:
            remsp = self.reynolds

        x = (remsp / 1e7) ** 0.3
        if 9e4 <= remsp < 1e7:
            upper_limit = 100**x
        elif 1e7 <= remsp:
            upper_limit = 100
        else:
            raise ValueError("Reynolds number out of specified range.")

        if 9e4 <= remsp < 1e6:
            lower_limit = 0.01**x
        elif 1e6 <= remsp:
            lower_limit = 0.1
        else:
            raise ValueError("Reynolds number out of specified range.")

        if lower_limit < self.reynolds_ratio < upper_limit:
            within_limits = True
        else:
            within_limits = False

        return {
            "lower": lower_limit,
            "upper": upper_limit,
            "within_limits": within_limits,
        }

    def plot_reynolds(self, fig=None, **kwargs):
        """Plot allowable Reynolds range and point.

        This will plot the allowable Mach range and the point according to the
        PTC criteria.

        Parameters
        ----------
        fig : plotly.Figure
            Plotly figure.

        Returns
        -------
        fig : plotly.Figure
            Plotly figure.
        """
        # build acceptable region
        upper_limit = []
        lower_limit = []
        remsp_range = np.geomspace(9e4, 1e9, 300)
        for remsp in remsp_range:
            reynolds_limits = self.reynolds_limits(remsp)
            lower_limit.append(reynolds_limits["lower"])
            upper_limit.append(reynolds_limits["upper"])

        fig = go.Figure()
        fig.add_trace(
            go.Scatter(x=remsp_range, y=lower_limit, marker=dict(color="black"))
        )
        fig.add_trace(
            go.Scatter(x=remsp_range, y=upper_limit, marker=dict(color="black"))
        )

        # add point
        fig.add_trace(
            go.Scatter(
                x=[self.reynolds.m], y=[self.reynolds_ratio], marker=dict(color="red")
            )
        )

        fig.update_xaxes(
            type="log",
            tickformat=".1e",
            tickmode="array",
            tickvals=[10**i for i in range(4, 9)],
            title=r"$\text{Machine Reynolds No. Specified} - Rem_{sp}$",
        )
        fig.update_yaxes(
            type="log",
            tickformat=".1e",
            tickmode="array",
            tickvals=[10**i for i in range(-3, 3)],
            title=r"$Rem_t / Rem_{sp}$",
        )
        fig.update_layout(showlegend=False)

        return fig

    def similarity_table(self, fig=None, **kwargs):
        """Plot similarity table.

        This table show the values for the non dimensional numbers (Mach, Reynolds
        and Volume ratio) and their calculated relations with respect to the
        original points used in the conversion (in the formulas, 'c' means converted
        points and 'o' means original point).

        If values are within limits, relation cells are colored in green, otherwise
        they are colored in red.

        """
        if fig is None:
            fig = go.Figure()

        quantity = ["Ratio of Specific Volume", "Mach Number", "Reynolds Number"]
        abbrev = ["$v_i / v_d$", "Mm", "Rem"]
        point_value = [
            f"{self.volume_ratio.m:.3f}",
            f"{self.mach.m:.3f}",
            f"{self.reynolds.m:.3e}",
        ]
        formula = [
            "$(v_i / v_d)_c / (v_i / v_d)_o = $",
            "$Mm_c - Mm_o = $",
            "$Rem_c / Rem_o = $",
        ]
        relation = [
            f"{self.volume_ratio_ratio.m:.3f}",
            f"{self.mach_diff.m:.3f}",
            f"{self.reynolds_ratio.m:.3e}",
        ]
        mach_limits = self.mach_limits()
        reynolds_limits = self.reynolds_limits()
        lower_limit = [
            0.95,
            f'{mach_limits["lower"]:.3f}',
            f'{reynolds_limits["lower"]:.3e}',
        ]
        upper_limit = [
            1.05,
            f'{mach_limits["upper"]:.3f}',
            f'{reynolds_limits["upper"]:.3e}',
        ]

        if 0.95 < self.volume_ratio_ratio < 1.05:
            volume_ratio_within_limits = True
        else:
            volume_ratio_within_limits = False

        light_green = "#D8F3DC"
        dark_green = "#2D6A4F"
        light_red = "#FFB3C1"
        dark_red = "#A4133C"

        rel_fill_color = []
        rel_font_color = []
        for status in [
            volume_ratio_within_limits,
            mach_limits["within_limits"],
            reynolds_limits["within_limits"],
        ]:
            if status is True:
                rel_fill_color.append(light_green)
                rel_font_color.append(dark_green)
            else:
                rel_fill_color.append(light_red)
                rel_font_color.append(dark_red)

        fig = go.Figure(
            data=[
                go.Table(
                    header=dict(
                        values=[
                            "<b>Quantity</b>",
                            "<b></b>",
                            "<b>Point Value</b>",
                            "<b></b>",
                            "<b>Relation</b>",
                            "<b>Lower Limit</b>",
                            "<b>Upper Limit</b>",
                        ],
                        line_color="white",
                        fill_color="white",
                        align="center",
                        font=dict(color="black", size=12),
                    ),
                    cells=dict(
                        values=[
                            quantity,
                            abbrev,
                            point_value,
                            formula,
                            relation,
                            lower_limit,
                            upper_limit,
                        ],
                        line_color=[
                            "white",
                            "white",
                            "white",
                            "white",
                            "white",
                            "white",
                            "white",
                        ],
                        fill_color=[
                            "white",
                            "white",
                            "white",
                            "white",
                            rel_fill_color,
                            "white",
                            "white",
                        ],
                        align="center",
                        font=dict(
                            color=[
                                "black",
                                "black",
                                "black",
                                "black",
                                rel_font_color,
                                "black",
                                "black",
                            ],
                            size=[12],
                        ),
                    ),
                )
            ]
        )

        return fig

    def plot_similarity(self, fig=None, **kwargs):
        """Plot similarity results.

        Plots the similarity results showing the Mach and Reynolds plots with
        their respective limits and also a table summarizing the results comparing
        the current (converted) point to the original point.

        Parameters
        ----------
        fig : plotly.Figure
            Plotly figure.

        Returns
        -------
        fig : plotly.Figure
            Plotly figure.
        """
        if fig is None:
            fig = make_subplots(
                rows=2,
                cols=2,
                specs=[
                    [{"type": "xy"}, {"type": "xy"}],
                    [{"type": "table", "colspan": 2}, None],
                ],
            )

        stable = self.similarity_table()
        mach = self.plot_mach()
        reynolds = self.plot_reynolds()

        for data in mach.data:
            data.showlegend = False
            fig.append_trace(data, row=1, col=1)

        for data in reynolds.data:
            data.showlegend = False
            fig.append_trace(data, row=1, col=2)

        for data in stable.data:
            fig.append_trace(data, row=2, col=1)

        fig.update_xaxes(mach.layout.xaxis, row=1, col=1)
        fig.update_xaxes(reynolds.layout.xaxis, row=1, col=2)
        fig.update_yaxes(mach.layout.yaxis, row=1, col=1)
        fig.update_yaxes(reynolds.layout.yaxis, row=1, col=2)

        return fig


def plot_func(self, attr):
    def inner(*args, plot_kws=None, **kwargs):
        """Plot parameter versus volumetric flow.

        You can choose units with the arguments flow_v_units='...' and
        attr_units='...'.
        """
        fig = kwargs.pop("fig", None)

        if fig is None:
            fig = go.Figure()

        if plot_kws is None:
            plot_kws = {}

        point_attr = r_getattr(self, attr)
        if callable(point_attr):
            point_attr = point_attr()

        flow_v_units = kwargs.get("flow_v_units", self.flow_v.units)
        # Split in '.' for cases such as disch.rho.
        # In this case the user gives rho_units instead of disch.rho_units
        attr_units = kwargs.get(f"{attr.split('.')[-1]}_units", point_attr.units)

        if attr_units is not None:
            point_attr = point_attr.to(attr_units)

        value = getattr(point_attr, "magnitude")
        units = getattr(point_attr, "units")

        flow_v = self.flow_v

        name = kwargs.get(
            "name", f"Flow: {flow_v.to(flow_v_units).m:.2f}, {attr}: {value:.2f}"
        )

        if flow_v_units is not None:
            flow_v = flow_v.to(flow_v_units)

        fig.add_trace(go.Scatter(x=[flow_v], y=[value], name=name, **plot_kws))

        return fig

    return inner


def n_exp(suc, disch):
    r"""Polytropic exponent.

    The polytropic exponent :math:`n` is calculated as per :cite:`schultz1962` eq. 27:

    .. math::

        \begin{equation}
            n = \frac{\log{\frac{p_d}{p_s}}}{\log{\frac{v_s}{v_d}}}
        \end{equation}

    Parameters
    ----------
    suc : ccp.State
        Suction state.
    disch : ccp.State
        Discharge state.

    Returns
    -------
    n_exp : float
        Polytropic exponent.
    """
    ps = suc.p()
    vs = 1 / suc.rho()
    pd = disch.p()
    vd = 1 / disch.rho()

    return np.log(pd / ps) / np.log(vs / vd)


def head_pol(suc, disch):
    r"""Polytropic head.

    The polytropic head is calculated as per :cite:`schultz1962` eq. 27:

    .. math::

       \begin{equation}
          H_p = (\frac{n}{n - 1}) (p_d v_d - p_s v_s)
       \end{equation}

    And :math:`n` is calculated by :py:func:`n_exp`.

    Parameters
    ----------
    suc : ccp.State
        Suction state.
    disch : ccp.State
        Discharge state.

    Returns
    -------
    head_pol : pint.Quantity
        Polytropic head (J/kg).
    """

    n = n_exp(suc, disch)

    p2 = disch.p()
    v2 = 1 / disch.rho()
    p1 = suc.p()
    v1 = 1 / suc.rho()

    return (n / (n - 1)) * (p2 * v2 - p1 * v1).to("joule/kilogram")


def eff_pol(suc, disch):
    """Polytropic efficiency.

    Parameters
    ----------
    suc : ccp.State
        Suction state.
    disch : ccp.State
        Discharge state.

    Returns
    -------
    eff_pol : pint.Quantity
        Polytropic efficiency (dimensionless).

    """
    wp = head_pol(suc, disch)

    dh = disch.h() - suc.h()

    return wp / dh


def head_isentropic(suc, disch):
    """Isentropic head.

    Parameters
    ----------
    suc : ccp.State
        Suction state.
    disch : ccp.State
        Discharge state.

    Returns
    -------
    head_isentropic : pint.Quantity
        Isentropic head.
    """
    # define state to isentropic discharge using dummy state
    disch_s = copy(disch)
    disch_s.update(p=disch.p(), s=suc.s())

    return head_pol(suc, disch_s).to("joule/kilogram")


def eff_isentropic(suc, disch):
    """Isentropic efficiency.

    Parameters
    ----------
    suc : ccp.State
        Suction state.
    disch : ccp.State
        Discharge state.

    Returns
    -------
    eff_isentropic : pint.Quantity
        Isentropic efficiency.
    """
    ws = head_isentropic(suc, disch)
    dh = disch.h() - suc.h()

    return ws / dh


def f_schultz(suc, disch):
    r"""Correction factor as per :cite:`schultz1962` eq 32:

    .. math::

       \begin{equation}
          f = \frac{H_{ds} - H_s}{(\frac{n_s}{n_s - 1})(p_d v_{ds} - p_s v_s)}
       \end{equation}


    Parameters
    ----------
    suc : ccp.State
        Suction state.
    disch : ccp.State
        Discharge state.

    Returns
    -------
    f_schultz : float
        Schultz polytropic factor.
    """

    # define state to isentropic discharge using dummy state
    disch_s = copy(disch)
    disch_s.update(p=disch.p(), s=suc.s())

    h2s_h1 = disch_s.h() - suc.h()
    h_isen = head_isentropic(suc, disch)

    return h2s_h1 / h_isen


def head_pol_schultz(suc, disch):
    r"""Polytropic head corrected by the :cite:`schultz1962` factor.

    .. math::

       \begin{equation}
          H_{p_{schultz}} = f_{schultz} H_p
       \end{equation}

    Where :math:`f_{schultz}` is calculated by :py:func:`f_schultz` and
    :math:`H_p` is calculated by :py:func:`head_pol`.

    Parameters
    ----------
    suc : ccp.State
        Suction state.
    disch : ccp.State
        Discharge state.

    Returns
    -------
    head_pol_schultz : pint.Quantity
        Schultz polytropic head (J/kg).
    """
    f = f_schultz(suc, disch)
    head = head_pol(suc, disch)

    return f * head


def eff_pol_schultz(suc, disch):
    """Polytropic efficiency as per :cite:`schultz1962`.

    Parameters
    ----------
    suc : ccp.State
        Suction state.
    disch : ccp.State
        Discharge state.

    Returns
    -------
    eff_pol_schultz : pint.Quantity
        Schultz polytropic efficiency (dimensionless).
    """
    wp = head_pol_schultz(suc, disch)
    dh = disch.h() - suc.h()

    return (wp / dh).to("dimensionless")


def head_pol_mallen_saville(suc, disch):
    r"""Polytropic head as per :cite:`mallen1977polytropic` calculated with:

    .. math::

       \begin{equation}
          H_p = (h_d - h_s) - (s_d - s_s) \frac{T_d - Ts}{\ln{(\frac{T_d}{T_s})}}
       \end{equation}

    Parameters
    ----------
    suc : ccp.State
        Suction state.
    disch : ccp.State
        Discharge state.

    Returns
    -------
    head_pol_mallen_saville : pint.Quantity
        Mallen-Saville polytropic polytropic head (J/kg).
    """

    head = (disch.h() - suc.h()) - (disch.s() - suc.s()) * (
        disch.T() - suc.T()
    ) / np.log(disch.T() / suc.T())

    return head


def eff_pol_mallen_saville(suc, disch):
    """Polytropic efficiency as per :cite:`mallen1977polytropic`.

    Parameters
    ----------
    suc : ccp.State
        Suction state.
    disch : ccp.State
        Discharge state.

    Returns
    -------
    eff_pol_mallen_saville : pint.Quantity
        Mallen-Saville polytropic efficiency (dimensionless).
    """
    wp = head_pol_mallen_saville(suc, disch)
    dh = disch.h() - suc.h()

    return (wp / dh).to("dimensionless")


_ref_H = 0


def head_reference(suc, disch, num_steps=100):
    r"""Reference head.

    The reference head consists of the integration of :math:`v dp` along the
    polytropic path as described by :cite:`huntington1985` and :cite:`sandberg2013limitations`.
    To achieve this we break the polytropic path into a series of subpaths.
    The compression ratio :math:`R_{c_i}` for each segment, as described by
    :cite:`sandberg2013limitations` is calculated with:

    .. math::

       \begin{equation}
          R_{c_i} = \sqrt[n_{steps}]{\frac{p_d}{p_s}}
       \end{equation}


    The calculation consists of two loops.
    One converges the :math:`T_1` temperature at each step by evaluating the
    difference between :math:`H = v_{avg} \Delta_p` and :math:`H = e \Delta_h`.
    The other evaluates the efficiency by checking the difference between
    the last :math:`T_1` to the discharge temperature :math:`T_d`.

    Parameters
    ----------
    suc : ccp.State
        Suction state.
    disch : ccp.State
        Discharge state.

    Returns
    -------
    head_reference : pint.Quantity
       Reference head as described by :cite:`huntington1985`. (J/kg).
    eff_reference : float
        Reference efficiency as described by :cite:`huntington1985` (dimensionless).
    """

    def calc_step_discharge_temp(T1, p1, p0, h0, v0, e):
        s1 = State.define(p=p1, T=T1, fluid=suc.fluid)
        h1 = s1.h()

        vm = (v0 + s1.v()) / 2
        delta_p = Q_(p1 - p0, "Pa")
        H0 = vm * delta_p
        H1 = e * (h1 - h0)

        return (H1 - H0).magnitude

    def calc_eff(e, suc, disch):
        rc = (disch.p().m / suc.p().m) ** (1 / num_steps)
        p_intervals = [suc.p().m]
        p = suc.p().m
        for i in range(num_steps):
            next_p = p * rc
            p = next_p
            p_intervals.append(p)

        T0 = suc.T().magnitude

        global _ref_H

        _ref_H = 0

        # TODO implement p_intervals considering pressure ratio
        for p0, p1 in zip(p_intervals[:-1], p_intervals[1:]):
            s0 = State.define(p=p0, T=T0, fluid=suc.fluid)
            T1 = newton(
                calc_step_discharge_temp, (T0 + 1e-3), args=(p1, p0, s0.h(), s0.v(), e)
            )
            s1 = State.define(p=p1, T=T1, fluid=suc.fluid)
            _ref_H += head_pol(s0, s1)

            T0 = T1

        return disch.T().magnitude - T1

    _ref_eff = newton(calc_eff, 0.8, args=(suc, disch))

    return _ref_H, _ref_eff


_ref_H_2017 = 0


def head_reference_2017(suc, disch, num_steps=100):
    r"""Reference head.

    The reference head consists of the integration along the
    polytropic path as described by :cite:`huntington2017`.
    Contrary to the method presented by :cite:`huntington1985`, this method does
    not use a specific volume linearized over each step of the integration, instead,
    it is based on the assumption that the compressibility factor varies linearly
    with the pressure within each step.

    In this case the inner loop of the method is calculated by:

    .. math::

       \begin{equation}
          a = \frac{z_i (\frac{p_{i+1}}{p_i}) - z_{i+1}}
          {(\frac{p_{i+1}}{p_i} - 1)} \\
          b = \frac{z_{i+1} - z_i}
          {(\frac{p_{i+1}}{p_i} - 1)} \\
          (s_{i+1} - s_i) = R \frac{(1-e)}{e}(a \ln{(\frac{p_{i+1}}{p_i})} + b(\frac{p_{i+1}}{p_i} - 1))
      \end{equation}


    Parameters
    ----------
    suc : ccp.State
        Suction state.
    disch : ccp.State
        Discharge state.

    Returns
    -------
    head_reference : pint.Quantity
       Reference head as described by :cite:`huntington2017`. (J/kg).
    eff_reference : float
        Reference efficiency as described by :cite:`huntington2017` (dimensionless).
    """
    R = suc.gas_constant() / suc.molar_mass()
    rc = (disch.p().m / suc.p().m) ** (1 / num_steps)
    p_intervals = [suc.p().m]
    p = suc.p().m
    for i in range(num_steps):
        next_p = p * rc
        p = next_p
        p_intervals.append(p)

    def calc_step_discharge_z(s1, s0, p1, p0, z0, R, e):
        state1 = ccp.State.define(p=p1, s=s1, fluid=suc.fluid)
        z1 = state1.z()
        a = (z0 * (p1 / p0) - z1) / ((p1 / p0) - 1)
        b = (z1 - z0) / ((p1 / p0) - 1)

        return (
            (R * ((1 - e) / e)) * (a * np.log(p1 / p0) + b * ((p1 / p0) - 1))
            - (state1.s() - Q_(s0, state1.s().units))
        ).magnitude

    def calc_eff(e, suc, disch, p_intervals):
        global _ref_H_2017
        _ref_H_2017 = 0
        s0 = suc.s().magnitude

        for p0, p1 in zip(p_intervals[:-1], p_intervals[1:]):
            state0 = ccp.State.define(p=p0, s=s0, fluid=suc.fluid)
            z0 = state0.z()

            s1 = newton(calc_step_discharge_z, (s0 + 1e-8), args=(s0, p1, p0, z0, R, e))
            state1 = ccp.State.define(p=p1, s=s1, fluid=suc.fluid)
            _ref_H_2017 += ccp.point.head_pol(state0, state1)

            s0 = s1
            T1 = state1.T().magnitude

        return disch.T().magnitude - T1

    eff0 = ccp.point.eff_pol_huntington(suc, disch)
    _ref_eff = newton(calc_eff, eff0, args=(suc, disch, p_intervals))

    return _ref_H_2017, _ref_eff


def f_sandberg_colby(suc, disch):
    r"""Correction factor as proposed by :cite:`sandberg2013limitations`.

    .. math::

       \begin{equation}
          f_p =
           \frac{(h_d - h_s) - T_{avg} (s_d - s_s)}
          {(\frac{n}{n-1})(p_d v_d - p_s v_s)}
       \end{equation}

    Parameters
    ----------
    suc : ccp.State
        Suction state.
    disch : ccp.State
        Discharge state.

    Returns
    -------
    f_sandberg_colby : pint.Quantity
       Polytropic head correction factor as described by :cite:`sandberg2013limitations` (dimensionless).
    """
    Tm = (suc.T() + disch.T()) / 2
    hd = disch.h()
    hs = suc.h()
    sd = disch.s()
    ss = suc.s()
    n = n_exp(suc, disch)
    pd = disch.p()
    ps = suc.p()
    vd = disch.v()
    vs = suc.v()

    f_sandberg_colby = ((hd - hs) - Tm * (sd - ss)) / (
        (n / (n - 1)) * (pd * vd - ps * vs)
    )

    return f_sandberg_colby.to("dimensionless")


def head_pol_sandberg_colby(suc, disch):
    r"""Polytropic head corrected by the :cite:`sandberg2013limitations` factor.

    .. math::

       \begin{equation}
          H_{p_{s-c}} = f_{s-c} H_p
       \end{equation}

    Where :math:`f_{s-c}` is calculated by :py:func:`f_sandberg_colby` and
    :math:`H_p` is calculated by :py:func:`head_pol`.

    Parameters
    ----------
    suc : ccp.State
        Suction state.
    disch : ccp.State
        Discharge state.

    Returns
    -------
    head_pol_sandberg_colby : pint.Quantity
       Reference head as described by :cite:`sandberg2013limitations` (J/kg).
    """
    f = f_sandberg_colby(suc, disch)
    h = f * head_pol(suc, disch)
    return h


def eff_pol_sandberg_colby(suc, disch):
    """Sandberg-Colby polytropic efficiency.

    Parameters
    ----------
    suc : ccp.State
        Suction state.
    disch : ccp.State
        Discharge state.

    Returns
    -------
    eff_pol_sandberg_colby: pint.Quantity
        Sandberg-Colby polytropic efficiency (dimensionless).
    """
    wp = head_pol_sandberg_colby(suc, disch)
    dh = disch.h() - suc.h()

    return (wp / dh).to("dimensionless")


def head_pol_huntington(suc, disch):
    r"""Polytropic head calculated by the 3 point method described by :cite:`huntington1985`.

    The polytropic head in this case is calculated from the polytropic efficiency with:

    .. math::

       \begin{equation}
          \frac{1}{e} =
          1 +
          \frac{\frac{(s_d - s_s)}{R}}
          {a \ln(\frac{p_d}{p_s})
          + b((\frac{p_d}{p_s}) - 1)
          + \frac{c}{2}(\ln{(\frac{p_d}{p_s})})^2}
       \end{equation}

    The constants :math:`a`, :math:`b` and :math:`c` are calculated with:

    .. math::

       \begin{equation}
          a = z_s - b \\
          b = \frac{(z_s + z_d - 2z_{int})}{((\frac{p_s}{p_s})^{0.5} - 1)^2} \\
          c = \frac{(z_d - a - b(\frac{p_d}{p_s}))}{\ln{(\frac{p_d}{p_s})}}
       \end{equation}

    The intermediate values are calculated interactively:

    .. math::

       \begin{equation}
          p_{int} = \sqrt{p_s p_d} \\
          T_{int}' = T_{int} \exp{(\frac{(s_{int}' - s_{int})}{c_p})}
       \end{equation}

    And :math:`s_{int}` is calculated by:

    .. math::

       \begin{equation}
            s_{int}' = s_s + (s_d - s_s)
            \frac{\frac{a}{2}\ln{(\frac{p_d}{p_s})} + b((\frac{p_s}{p_s})^{0.5} - 1)) + \frac{c}{8}(\ln{(\frac{p_d}{p_s})})^2}
            {a\ln(\frac{p_d}{p_s}) + b((\frac{p_d}{p_s})-1) + \frac{c}{2}(\ln(\frac{p_d}{p_s}))^2}
       \end{equation}

    Parameters
    ----------
    suc : ccp.State
        Suction state.
    disch : ccp.State
        Discharge state.

    Returns
    -------
    head_pol_huntington : pint.Quantity
       Polytropic head as described by :cite:`huntington1985` (J/kg).
    """
    eff = eff_pol_huntington(suc, disch)
    head = (disch.h() - suc.h()) * eff

    return head


def eff_pol_huntington(suc, disch):
    """Polytropic efficiency calculated by the 3 point method described by :cite:`huntington1985`.

    Parameters
    ----------
    suc : ccp.State
        Suction state.
    disch : ccp.State
        Discharge state.

    Returns
    -------
    eff_pol_huntington : pint.Quantity
       Polytropic efficiency as described by :cite:`huntington1985` (dimensionless).
    """
    p1 = suc.p()
    p2 = disch.p()
    s1 = suc.s()
    s2 = disch.s()
    z1 = suc.z()
    z2 = disch.z()
    T1 = suc.T()
    T2 = disch.T()
    p3 = np.sqrt(p1 * p2)

    T3 = np.sqrt(T1 * T2)
    error = 1
    n = 0
    while error > 1e-10:
        state3 = State.define(p=p3, T=T3, fluid=suc.fluid)
        s3 = state3.s()
        z3 = state3.z()
        cp3 = state3.cp()
        b = (z1 + z2 - 2 * z3) / (np.sqrt(p2 / p1) - 1) ** 2
        a = z1 - b
        c = (z2 - a - b * (p2 / p1)) / np.log(p2 / p1)
        s3_ = s1 + (s2 - s1) * (
            (
                ((a / 2) * np.log(p2 / p1))
                + b * (np.sqrt(p2 / p1) - 1)
                + (c / 8) * np.log(p2 / p1) ** 2
            )
            / (
                a * np.log(p2 / p1)
                + b * ((p2 / p1) - 1)
                + (c / 2) * np.log(p2 / p1) ** 2
            )
        )
        T3_new = T3 * np.exp((s3_ - s3) / cp3)
        error = abs(T3_new - T3).m
        T3 = T3_new

        n += 1
        if n == 100:
            raise RecursionError("Maximum number of iterations exceeded.")

    R = suc.gas_constant() / suc.molar_mass()
    inv_e = 1 + (
        ((s2 - s1) / R)
        / (a * np.log(p2 / p1) + b * ((p2 / p1) - 1) + (c / 2) * np.log(p2 / p1) ** 2)
    )
    eff = 1 / inv_e

    return eff


@check_units
def power_calc(flow_m, head, eff):
    """Calculate power.

    Parameters
    ----------
    flow_m : pint.Quantity, float
        Mass flow (kg/s).
    head : pint.Quantity, float
        Head (J/kg).
    eff : pint.Quantity, float
        Efficiency (dimensionless).

    Returns
    -------
    power : pint.Quantity
        Power (watt).
    """
    power = flow_m * head / eff

    return power.to("watt")


@check_units
def u_calc(D, speed):
    """Calculate the impeller tip speed.

    Parameters
    ----------
    D : pint.Quantity, float
        Impeller diameter (m).
    speed : pint.Quantity, float
        Impeller speed (rad/s).

    Returns
    -------
    u_calc : pint.Quantity
        Impeller tip speed (m/s).
    """
    u = speed * D / 2
    return u.to("m/s")


@check_units
def psi(head, speed, D):
    """Polytropic head coefficient.

    Parameters
    ----------
    head : pint.Quantity, float
        Polytropic head (J/kg).
    speed : pint.Quantity, float
        Impeller speed (rad/s).
    D : pint.Quantity, float
        Impeller diameter.

    Returns
    -------
    psi : pint.Quantity
        Polytropic head coefficient (dimensionless).
    """
    u = u_calc(D, speed)
    psi = head / (u**2 / 2)
    return psi.to("dimensionless")


@check_units
def u_from_psi(head, psi):
    """Calculate u_calc from non dimensional psi.

    Parameters
    ----------
    head : pint.Quantity, float
        Polytropic head.
    psi : pint.Quantity, float
        Head coefficient.

    Returns
    -------
    u_calc : pint.Quantity, float
        Impeller tip speed.
    """
    u = np.sqrt(2 * head / psi)

    return u.to("m/s")


@check_units
def speed_from_psi(D, head, psi):
    """Calculate speed from non dimensional psi.

    Parameters
    ----------
    D : pint.Quantity, float
        Impeller diameter.
    head : pint.Quantity, float
        Polytropic head.
    psi : pint.Quantity, float
        Head coefficient.

    Returns
    -------
    u_calc : pint.Quantity, float
        Impeller tip speed.
    """
    u = u_from_psi(head, psi)

    speed = 2 * u / D

    return speed.to("rad/s")


@check_units
def phi(flow_v, speed, D):
    """Flow coefficient."""
    u = u_calc(D, speed)

    phi = flow_v * 4 / (np.pi * D**2 * u)

    return phi.to("dimensionless")


@check_units
def flow_from_phi(D, phi, speed):
    """Calculate flow from non dimensional phi.

    Parameters
    ----------
    D : pint.Quantity, float
        Impeller diameter (m).
    phi : pint.Quantity, float
        Flow coefficient (m³/s).
    speed : pint.Quantity, float
        Speed (rad/s).

    Returns
    -------
    u_calc : pint.Quantity, float
        Impeller tip speed.
    """
    u = speed * D / 2

    flow_v = phi * (np.pi * D**2 * u) / 4

    return flow_v.to("m**3/s")


def head_from_psi(D, psi, speed):
    """Calculate head from non dimensional psi.

    Parameters
    ----------
    D : pint.Quantity, float
        Impeller diameter (m).
    psi : pint.Quantity, float
        Head coefficient.
    speed : pint.Quantity, float
        Speed (rad/s).

    Returns
    -------
    u_calc : pint.Quantity, float
        Impeller tip speed.
    """
    u = speed * D / 2
    head = psi * (u**2 / 2)

    return head.to("J/kg")


def disch_from_suc_head_eff(suc, head, eff, polytropic_method=None):
    """Calculate discharge state from suction, head and efficiency.

    Parameters
    ----------
    suc : ccp.State
        Suction state.
    head : pint.Quantity, float
        Polytropic head (J/kg).
    eff : pint.Quantity, float
        Polytropic efficiency (dimensionless).

    Returns
    -------
    disch : ccp.State
        Discharge state.
    """
    if polytropic_method is None:
        polytropic_method = ccp.config.POLYTROPIC_METHOD

    head_calc_func = globals()[f"head_pol_{polytropic_method}"]
    h_disch = head / eff + suc.h()

    #  consider first an isentropic compression
    disch = State.define(h=h_disch, s=suc.s(), fluid=suc.fluid)

    def update_pressure(p):
        disch.update(h=h_disch, p=p)
        new_head = head_calc_func(suc, disch)

        return (new_head - head).magnitude

    newton(update_pressure, disch.p().magnitude, tol=1e-1)

    return disch


def disch_from_suc_disch_p_eff(suc, disch_p, eff, polytropic_method=None):
    """Calculate discharge state from suction, discharge pressure and efficiency.

    Parameters
    ----------
    suc : ccp.State
        Suction state.
    disch_p : pint.Quantity, float
        Discharge pressure (Pa).
    eff : pint.Quantity, float
        Polytropic efficiency (dimensionless).

    Returns
    -------
    disch : ccp.State
        Discharge state.
    """
    # consider first an isentropic compression
    if polytropic_method is None:
        polytropic_method = ccp.config.POLYTROPIC_METHOD

    disch = ccp.State.define(p=disch_p, s=suc.s(), fluid=suc.fluid)
    eff_calc_func = globals()[f"eff_pol_{polytropic_method}"]

    def update_state(x):
        disch.update(p=disch_p, T=x)
        new_eff = eff_calc_func(suc, disch)

        return (new_eff - eff).magnitude

    newton(update_state, disch.T().magnitude)

    return disch


@check_units
def reynolds(suc, speed, b, D):
    """Calculate the Reynolds number.

    Parameters
    ----------
    suc : ccp.State
        Suction state.
    speed : pint.Quantity, float
        Impeller speed (rad/s).
    b : pint.Quantity, float
        Impeller width (m).
    D : pint.Quantity, float
        Impeller diameter (m).

    Returns
    -------
    reynolds : pint.Quantity
        Reynolds number (dimensionless).
    """
    u = u_calc(D, speed)
    re = u * b * suc.rho() / suc.viscosity()

    return re.to("dimensionless")


@check_units
def mach(suc, speed, D):
    """Calculate the Mach number.

    Parameters
    ----------
    suc : ccp.State
        Suction state.
    speed : pint.Quantity, float
        Impeller speed (rad/s).
    D : pint.Quantity, float
        Impeller diameter (m).

    Returns
    -------
    mach : pint.Quantity
        Mach number (dimensionless).
    """
    u = u_calc(D, speed)
    a = suc.speed_sound()
    ma = u / a

    return ma.to("dimensionless")
